Package: optimalFlow
Type: Package
Title: optimalFlow
Version: 1.24.0
Author: Hristo Inouzhe <hristo.inouzhe@gmail.com>
Maintainer: Hristo Inouzhe <hristo.inouzhe@gmail.com>
Description: Optimal-transport techniques applied to supervised flow cytometry gating.
License: Artistic-2.0
Encoding: UTF-8
Depends: dplyr, optimalFlowData, rlang (>= 0.4.0)
Imports: transport, parallel, Rfast, robustbase, dbscan, randomForest,
        foreach, graphics, doParallel, stats, flowMeans, rgl, ellipse
Suggests: knitr, BiocStyle, rmarkdown, magick
VignetteBuilder: knitr
biocViews: Software, FlowCytometry, Technology
RoxygenNote: 7.1.0
git_url: https://git.bioconductor.org/packages/optimalFlow
git_branch: RELEASE_3_23
git_last_commit: e55f2b3
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 23:47:49 UTC; biocbuild

# Generated by roxygen2: do not edit by hand

export(costWasserMatchingEllipse)
export(cytoPlot)
export(cytoPlot3d)
export(cytoPlotDatabase)
export(cytoPlotDatabase3d)
export(estimCovCellGeneral)
export(estimationCellBarycenter)
export(f1Score)
export(f1ScoreVoting)
export(labelTransfer)
export(labelTransferEllipse)
export(optimalFlowClassification)
export(optimalFlowTemplates)
export(qdaClassification)
export(tclustWithInitialization)
export(tclust_H)
export(trimmedKBarycenter)
export(voteLabelTransfer)
export(w2dist)
export(wasserCostFunction)
import(dplyr)
import(ellipse)
import(optimalFlowData)
import(rgl)
import(rlang)
importFrom(Rfast,dista)
importFrom(Rfast,rowMaxs)
importFrom(flowMeans,flowMeans)
importFrom(foreach,"%:%")
importFrom(foreach,"%dopar%")
importFrom(graphics,lines)
importFrom(stats,as.dist)
importFrom(stats,cov)
importFrom(stats,cutree)
importFrom(stats,hclust)
importFrom(stats,mahalanobis)
importFrom(stats,predict)
importFrom(stats,rnorm)
importFrom(stats,runif)

#' distGaussianMean
#'
#' Computes the componente relative to the means of the squared 2-Wasserstein distance.
#'
#' @param N1 A multivariate normal distribution as list with elements mean and cov.
#' @param N2 A multivariate normal distribution as list with elements mean and cov.
#'
#' @return A double equivalent to the squared Euclidean distance between the means.
#'
#' @examples
#' distGaussianMean(list(mean = c(-1, -1), cov = diag(1, 2)), list(mean = c(1, 1), cov = diag(1, 2)))
#'
#' @noRd
#'
distGaussianMean <- function(N1, N2) {
    di.mean <- sum((N1$mean - N2$mean)^2)
    distancia <- di.mean
    distancia
}
#' distGaussianCov
#'
#' Computes the componente relative to the covariances of the squared 2-Wasserstein distance.
#'
#' @param N1 A multivariate normal distribution as list with elements mean and cov.
#' @param N2 A multivariate normal distribution as list with elements mean and cov.
#'
#' @return A double representing the squared 2-Wassersteindistance between normals with the same mean.
#'
#' @examples
#' distGaussianCov(list(mean = c(1, 1), cov = diag(2, 2)), list(mean = c(1, 1), cov = diag(1, 2)))
#'
#' @noRd
#'
distGaussianCov <- function(N1, N2) {
    A <- N1$cov
    B <- N2$cov
    sqrt.matrix <- function(C) {
        e <- eigen(C)
        e$values <- abs(e$values)
        sqrtA <- e$vectors %*% diag(sqrt(e$values)) %*% t(e$vectors)
        sqrtA
    }
    sqrtA <- sqrt.matrix(A)
    M1 <- sqrtA %*% B %*% sqrtA
    M2 <- sqrt.matrix(M1)
    M3 <- A + B - 2 * M2
    di.cov <- sum(diag(M3))
    distancia <- di.cov
    distancia
}
#' distGaussian
#'
#' Computes the squared 2-Wasserstein distance.
#'
#' @param N1 A multivariate normal distribution as list with elements mean and cov.
#' @param N2 A multivariate normal distribution as list with elements mean and cov.
#'
#' @return A double representing the squared 2-Wasserstein distance between normals.
#'
#' @examples
#' distGaussian(list(mean = c(-1, -1), cov = diag(2, 2)), list(mean = c(1, 1), cov = diag(1, 2)))
#'
#' @noRd
#'
distGaussian <- function(N1, N2) {
    di.mean <- sum((N1$mean - N2$mean)^2)
    A <- N1$cov
    B <- N2$cov
    sqrt.matrix <- function(C) {
        e <- eigen(C)
        e$values <- abs(e$values)
        sqrtA <- e$vectors %*% diag(sqrt(e$values)) %*% t(e$vectors)
        sqrtA
    }
    sqrtA <- sqrt.matrix(A)
    M1 <- sqrtA %*% B %*% sqrtA
    M2 <- sqrt.matrix(M1)
    M3 <- A + B - 2 * M2
    di.cov <- sum(diag(M3))
    distancia <- di.mean + di.cov
    distancia
}
#' GaussianBarycenters
#'
#' Calculates the (1-)barrycenter of a mixture of normals with the same mean.
#'
#' @param matrices A list of covariance matrices.
#' @param weight A vector of weights associated to the covariance matrices.
#'
#' @return A list formed by:
#' \describe{
#'  \item{Barycenter}{Returns the barycenter, i.e., a covariance matrix.}
#'  \item{Variation}{A double representing the Wasserstein Variation.}
#'  \item{Num.iter}{The number of iterations to achieve the stopping criteria.}
#' }
#'
#' @examples
#' GaussianBarycenters(list(diag(2, 2),diag(1, 2)), c(0.5, 0.5))
#'
#' @noRd
#'
GaussianBarycenters <- function(matrices, weight) {
    d <- ncol(matrices[[1]])
    k <- length(weight)
    Sn <- diag(d)
    trfix <- 0
    for (j in 1:k) trfix <- trfix + sum(weight[j] * sum(diag(matrices[[j]])))
    main.iter <- function(S) {
        VS <- sum(diag(S)) + trfix
        aux1 <- eigen(S)
        aux1$values <- abs(aux1$values)
        S12 <- aux1$vectors %*% diag(aux1$values^{
            1/2
        }) %*% t(aux1$vectors)
        S12inv <- aux1$vectors %*% diag(aux1$values^{
            -1/2
        }) %*% t(aux1$vectors)
        Saux <- list()
        trmen <- 0
        for (j in 1:k) {
            temp <- eigen(S12 %*% matrices[[j]] %*% S12)
            temp$values <- abs(temp$values)
            Saux[[j]] <- temp$vectors %*% diag(temp$values^{
                1/2
            }) %*% t(temp$vectors)
            Saux[[j]] <- weight[j] * Saux[[j]]
            trj <- sum(diag(Saux[[j]]))
            trmen <- trmen + trj
        }
        VS <- VS - 2 * trmen
        S <- Reduce("+", Saux)
        list(iterant = S, V = VS)
    }
    nuevo <- main.iter(Sn)
    Vold <- nuevo$V
    Sn <- nuevo$iterant
    tol <- 5
    n.iter <- 1
    while (tol > 1e-09) {
        nuevo <- main.iter(Sn)
        n.iter <- n.iter + 1
        Vnew <- nuevo$V
        tol <- Vold - Vnew
        Vold <- Vnew
        Sn <- nuevo$iterant
    }
    list(Barycenter = Sn, Variation = Vnew, Num.iter = n.iter)
}
#' kcenter
#'
#' Calculates the k-barycenter of a list of multivaraite normals for a given
#'  number of clusters ,k, and an assignation of each normal to the respective group
#'
#' @param points A list of multivariate normals, where each element is a list with values mean and cov.
#' @param kk The number k of groups for the k-barycenter.
#' @param center.asigned A vector indicating to which group each normal should be assigned.
#'
#' @return A list with values:
#' \describe{
#'  \item{kcenters}{A list of the elements of the k-barycenter, i.e., a list of k lists containing means and covariances.}
#'  \item{t.variation}{A double giving the trimmed wasserstein variation.}
#' }
#'
#' @examples
#' normals <- list(list(mean = c(1, 1), cov = diag(2, 2)), list(mean = c(1, 1), cov = diag(1, 2)),
#'  list(mean = c(3, 3), cov = diag(1, 2)))
#' kcenter(normals, 2, c(1, 1, 2))
#'
#' @noRd
#'
kcenter <- function(points, kk, center.asigned) {
    new.bar <- list()
    Trimmed.Variation.Prov <- 0
    for (a in 1:kk) {
        cluster.a <- points[which(center.asigned == a)]
        n.a <- length(cluster.a)
        means.a <- NULL
        covs.a <- list()
        for (i in 1:n.a) {
            means.a <- cbind(means.a, cluster.a[[i]]$mean)
            covs.a[[i]] <- cluster.a[[i]]$cov
        }
        mean.cluster.a <- apply(means.a, 1, mean)
        var.means.cluster.a <- mean(apply(means.a^2, 2, sum)) - sum(mean.cluster.a^2)
        Alg.Baricenter <- GaussianBarycenters(covs.a, weight = rep(1, n.a)/n.a)
        cov.cluster.a <- Alg.Baricenter$Barycenter
        new.bar[[a]] <- list(mean = mean.cluster.a, cov = cov.cluster.a)
        Trimmed.Variation.Prov <- Trimmed.Variation.Prov + Alg.Baricenter$Variation + var.means.cluster.a
    }
    resultado <- list(kcenters = new.bar, t.variation = Trimmed.Variation.Prov)
    resultado
}
#' wasserMinDist
#'
#' For two lists of multivariate normals calcualtes the closest member,
#'  in 2-Wasserstein distance, of the later list to each element of the former.
#'
#' @param points List of multivariate normals, where each element is a list with values mean and cov.
#' @param centres List of multivariate normals, where each element is a list with values mean and cov.
#'
#' @return A matrix where each column indicates the distance between the respective entry in points and
#'  the closest element in centers and the index of this closest element.
#'
#' @examples
#' normals <- list(list(mean = c(1, 1), cov = diag(2, 2)), list(mean = c(1, 1), cov = diag(1, 2)),
#'  list(mean = c(3, 3), cov = diag(1, 2)))
#' k_barycenter <- kcenter(normals, 2, c(1, 1, 2))$kcenters
#' wasserMinDist(normals, k_barycenter)
#'
#' @noRd
#'
wasserMinDist <- function(points, centres) {
    uu <- length(points)
    kk <- length(centres)
    aux.f1 <- function(u) {
        N1 <- points[[u]]
        aux.f2 <- function(j) distGaussian(N1, centres[[j]])
        distancias <- sapply(1:kk, aux.f2)
        resultado <- c(min(distancias), which.min(distancias))
        resultado
    }
    resultado <- sapply(1:uu, aux.f1)
    resultado
}
#' trimmedMinDist
#'
#' For two lists of multivaraite normals, points and centers, returns the index of which element
#'  in centers is closest to each element in points when some trimming is allowed.
#'
#' @param points List of multivariate normals, where each element is a list with values mean and cov.
#' @param centres List of multivariate normals, where each element is a list with values mean and cov.
#' @param alpha Level of triming.
#'
#' @return A vector with the index of which element in centers is closest to each element in points.
#'
#' @examples
#' normals <- list(list(mean = c(1, 1), cov = diag(2, 2)),list(mean = c(1, 1), cov = diag(1, 2)),
#'  list(mean = c(3, 3), cov = diag(1, 2)))
#' k_barycenter <- kcenter(normals, 2, c(1, 1, 2))$kcenters
#' trimmedMinDist(normals, k_barycenter, 0)
#'
#' @noRd
#'
trimmedMinDist <- function(points, centres, alpha = 0.1) {
    asignacion <- wasserMinDist(points, centres)
    orden.dista <- order(asignacion[1, ])
    threshold <- sort(asignacion[1, ])[floor(length(points) * (1 - alpha))]
    asignacion[2, (asignacion[1, ] > threshold)] <- 0
    resultado <- asignacion[2, ]
    resultado
}

#' voteTransformation
#'
#' Transforming votes obtained by using optimalFlowTemplates + OptimalFlowClassification with consenus.method in c('hierarchical', 'k-barycenter') and classif.method = 'matching' and cost.function = 'ellipses' to an appropriate format for using f1ScoreVoting.
#'
#' @param vote.0 Values obtained by voteLabelTransfer.
#' @param vote.1 Original proportions of the clusters after the template obtention.
#'
#' @return A list for the votes on each cell.
#'
#' @examples
#' vote.0 <- list('1' = data.frame(cell = c(1, 2), 'compound.proportion' = c(0.7, 0.3),
#'                                'simple.proportion'= c(0.7, 0.3)), '2' = data.frame(cell = c(1, 2),
#'                                                                                   'compound.proportion' = c(0.3, 0.7), 'simple.proportion'= c(0.3, 0.7)))
#' vote.1.1 <- t(c(0.8, 0.2))
#' names(vote.1.1) <- c('A', 'B')
#' vote.1.2 <- t(c(0.2, 0.8))
#' names(vote.1.2) <- c('A', 'B')
#' vote.1 <- list(vote.1.1, vote.1.2)
#' voteTransformation(vote.0, vote.1)
#'
#' @noRd
#'
voteTransformation <- function(vote.0, vote.1) {
    final.vote <- lapply(names(vote.0), function(i) {
        cells.0 <- as.character(vote.0[[i]]$cell)
        cells.1 <- vote.1[as.integer(cells.0)]
        joint.data.frame <- data.frame()
        joint.data.frame.1 <- data.frame()
        for (j in 1:length(cells.1)) {
            temp.data.frame <- data.frame(cell = names(cells.1[[j]]), simple.proportion = vote.0[[i]]$simple.proportion[j] * as.vector(cells.1[[j]]), 
                compound.proportion = vote.0[[i]]$compound.proportion[j] * as.vector(cells.1[[j]]))
            joint.data.frame = rbind(joint.data.frame, temp.data.frame)
        }
        cell.vote <- joint.data.frame %>% group_by(cell) %>% summarise(simple.proportion = sum(simple.proportion), compound.proportion = sum(compound.proportion))
        cell.vote$simple.proportion <- cell.vote$simple.proportion/sum(cell.vote$simple.proportion)
        cell.vote$compound.proportion <- cell.vote$compound.proportion/sum(cell.vote$compound.proportion)
        cell.vote <- as.data.frame(dplyr::arrange(cell.vote, desc(compound.proportion)))
    })
    names(final.vote) <- names(vote.0)
    return(final.vote)
}

#' voteLabelTransfer
#'
#' A wrapper for doing either labelTransfer or labelTransferEllipse.
#'
#' @param type 'points' indicates use of labelTransfer; 'ellipses' of labelTransferEllipse.
#' @param test.partition Only when type = 'points'. Labels of a partition of the test data.
#' @param test.cytometry Only when type = 'points'. Test data, a dataframe without labels.
#' @param test.partition.ellipse Only when type = 'ellipses'. A test clustering viewed as a mixture of multivariate normal distributions.
#' @param training.cytometries Only when type = 'points'. List of partitions, where each partition is a dataframe wher the last column contains the labels of the partition.
#' @param training.cytometries.barycenter Only when type = 'ellipses'. A training partition viewed as a mixture of multivariate normal distributions.
#' @param test Only when type = 'ellipses'. A dummy variable, should be any integral. Ment for use with lapply.
#' @param op.syst Type of system, takes values in c('unix', 'windows').
#' @param cl.paral Number of cores to be used in parallel procedures.
#' @param equal.weights If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.
#'
#' @return A list containing:
#' \describe{
#'  \item{final.vote}{A list for the votes on each cell.}
#'  \item{complete.vote}{A more complete list for the votes on each cell.}
#' }
#'
#' @examples
#' \donttest{
#' data.example <- data.frame(v1 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)),
#'                           v2 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)), id = c(rep(0, 50), rep(1, 50)))
#' test.labels <- c(rep('a', 50), rep('b', 50))
#' voteLabelTransfer(test.partition = test.labels, test.cytometry = data.example[, 1:2],
#'                   training.cytometries = list(data.example), op.syst = .Platform$OS.type)$final.vote[[1]]
#' }
#' @export
#'
voteLabelTransfer <- function(type = "points", test.partition, test.cytometry, test.partition.ellipse, training.cytometries, training.cytometries.barycenter,
    test = 1, op.syst, cl.paral = 1, equal.weights = FALSE) {
    if (type == "points") {
        transported.labels <- list()
        if (op.syst == "windows") {
            cl <- parallel::makePSOCKcluster(cl.paral)
            for (j in test) {
                t0 <- Sys.time()
                transported.labels[[as.character(j)]] <- parallel::parLapply(cl, training.cytometries, optimalFlow::labelTransfer,
                  test.cytometry, test.partition, equal.weights)
                t1 <- Sys.time()
                print(paste(t1 - t0, units(difftime(t1, t0))))
            }
            parallel::stopCluster(cl)
        } else {
            if (op.syst == "unix") {
                for (j in test) {
                  t0 <- Sys.time()
                  transported.labels[[as.character(j)]] <- parallel::mclapply(training.cytometries, optimalFlow::labelTransfer,
                    test.cytometry, test.partition, equal.weights, mc.cores = cl.paral)
                  t1 <- Sys.time()
                  print(paste(t1 - t0, units(difftime(t1, t0))))
                }

            }
        }
    } else {
        if (type == "ellipses") {
            test.partition <- 1:length(test.partition.ellipse)
            training.cytometries <- 1
            transported.labels <- list()
            for (j in test) {
                t0 <- Sys.time()
                transported.labels[[as.character(j)]] <- lapply(1, optimalFlow::labelTransferEllipse, test.partition.ellipse, training.cytometries.barycenter,
                  equal.weights)
                t1 <- Sys.time()
                print(paste(t1 - t0, units(difftime(t1, t0))))
            }
        }
    }

    vote.transport.labels <- list()
    complete.vote.transport.labels <- list()
    for (i in as.character(test)) {
        transp_lab_cito <- list()
        transp_lab_cito_complete <- list()
        for (test.cell in unique(transported.labels[[i]][[1]]$to)) {
            vote.on.test.cell <- c()
            for (j in 1:length(training.cytometries)) {
                receiving.cell.indexes <- which(transported.labels[[i]][[j]]$to == test.cell)
                sending.cells <- transported.labels[[i]][[j]]$from[receiving.cell.indexes]
                sending.cells.vote.proportion <- transported.labels[[i]][[j]]$mass[receiving.cell.indexes]/sum(transported.labels[[i]][[j]]$mass[receiving.cell.indexes])
                sending.cells.proportions <- array(dim = length(receiving.cell.indexes))
                t <- 0
                for (jj in receiving.cell.indexes) {
                  t <- t + 1
                  receiving.cell.indexes_jj <- which(transported.labels[[i]][[j]]$from == transported.labels[[i]][[j]]$from[jj])
                  sending.cells.proportions[t] <- transported.labels[[i]][[j]]$mass[jj]/sum(transported.labels[[i]][[j]]$mass[receiving.cell.indexes_jj])
                }
                vote.on.test.cell <- rbind(vote.on.test.cell, cbind(rep(j, t), sending.cells, sending.cells.vote.proportion, sending.cells.proportions))
            }
            vote.on.test.cell <- data.frame(id = as.integer(unlist(vote.on.test.cell[, 1])), cell = vote.on.test.cell[, 2], vote.proportion = as.double(unlist(vote.on.test.cell[,
                3])), original.proportion = as.double(unlist(vote.on.test.cell[, 4])))
            vote.on.test.cell$compound.proportion <- vote.on.test.cell$vote.proportion * vote.on.test.cell$original.proportion
            vote.on.test.cell.bis <- as.data.frame(vote.on.test.cell %>% group_by(cell) %>% summarise(compound.proportion = sum(compound.proportion),
                simple.proportion = sum(vote.proportion)))
            vote.on.test.cell.bis$compound.proportion <- vote.on.test.cell.bis$compound.proportion/sum(vote.on.test.cell.bis$compound.proportion)
            vote.on.test.cell.bis$simple.proportion <- vote.on.test.cell.bis$simple.proportion/sum(vote.on.test.cell.bis$simple.proportion)
            transp_lab_cito[[test.cell]] <- vote.on.test.cell.bis[order(-vote.on.test.cell.bis$compound.proportion), ]
            transp_lab_cito_complete[[test.cell]] <- vote.on.test.cell
        }
        vote.transport.labels[[i]] <- transp_lab_cito
        complete.vote.transport.labels[[i]] <- transp_lab_cito_complete
    }
    return(list(final.vote = vote.transport.labels, complete.vote = complete.vote.transport.labels))
}

#' cytoPlotDatabase3d
#'
#' A plot3d wrapper for a database (list) of cytometries as a mixture of multivariate normals as used in optimalFlowTemplates.
#'
#' @param database.cytometries.as.mixtures A list where each component is a mixture distribution. That is, each component is a list, where each element contains the parameters of a component of the mixture as a list with entries: mean, cov, weight and type.
#' @param dimensions A vector containing the two variables on which to perform the projection.
#' @param xlim the x limits (x1, x2) of the plot. Note that x1 > x2 is allowed and leads to a ‘reversed axis’. The default value, NULL, indicates that the range of the finite values to be plotted should be used.
#' @param ylim the y limits of the plot.
#' @param zlim the z limits of the plot.
#' @param xlab a label for the x axis, defaults to a description of x.
#' @param ylab a label for the y axis, defaults to a description of y.
#' @param zlab a label for the z axis, defaults to a description of z.
#' @param colour If TRUE plots elements of a mixture distribution in different colours. If FALSE plots them in black.
#'
#' @return A three dimensional plot of ellipsoids containing the 95% of probability for the respectives components of the mixture distributions.
#' @examples
#' database <- buildDatabase(
#'   dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
#'   population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
#' templates.optimalFlow <-
#'   optimalFlowTemplates(
#'     database = database, templates.number = 5, cl.paral = 1
#'   )
#' # # To execute requires an actual monitor since it uses rgl.
#' # cytoPlotDatabase3d(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000), xlab = "", ylab = "", ylab = c(0, 8000))
#'
#' @import rgl
#' @export
cytoPlotDatabase3d <- function (database.cytometries.as.mixtures, dimensions = c(1, 2, 3), xlim = c(0, 8000),
                                ylim = c(0, 8000), zlim = c(0, 8000), xlab = "", ylab = "", zlab = "", colour = TRUE){
  n.cytometries <- length(database.cytometries.as.mixtures)
  cytometry.as.mixture <- database.cytometries.as.mixtures[[1]]
  n.cells <- length(cytometry.as.mixture)
  rgl::plot3d(rgl::ellipse3d(cytometry.as.mixture[[1]]$
                               cov[dimensions, dimensions],
                             centre = cytometry.as.mixture[[1]]$
                               mean[dimensions]),
              xlim = xlim, ylim = ylim, zlim = zlim, alpha = 0.5,
              col = 1, xlab = xlab, ylab = ylab,
              zlab = zlab)
  if (colour) {
    for (j in 2:n.cells){
      rgl::plot3d(rgl::ellipse3d(cytometry.as.mixture[[j]]$cov[dimensions, dimensions],
                                 centre = cytometry.as.mixture[[j]]$
                                   mean[dimensions]), alpha = 0.5, add = TRUE, col = j)
    }
    for (i in 2:n.cytometries){
      cytometry.as.mixture <- database.cytometries.as.mixtures[[i]]
      n.cells <- length(cytometry.as.mixture)
      for (j in 1:n.cells){
        rgl::plot3d(rgl::ellipse3d(cytometry.as.mixture[[j]]$cov[dimensions, dimensions],
                                   centre = cytometry.as.mixture[[j]]$
                                     mean[dimensions]), alpha = 0.5, add = TRUE, col = j)
      }
    }
  } else {
    for (j in 2:n.cells){
      rgl::plot3d(rgl::ellipse3d(cytometry.as.mixture[[j]]$cov[dimensions, dimensions],
                                 centre = cytometry.as.mixture[[j]]$
                                   mean[dimensions]), alpha = 0.5, add = TRUE, col = 1)
    }
    for (i in 2:n.cytometries){
      cytometry.as.mixture <- database.cytometries.as.mixtures[[i]]
      n.cells <- length(cytometry.as.mixture)
      for (j in 1:n.cells){
        rgl::plot3d(rgl::ellipse3d(cytometry.as.mixture[[j]]$cov[dimensions, dimensions],
                                   centre = cytometry.as.mixture[[j]]$
                                     mean[dimensions]), alpha = 0.5, add = TRUE, col = 1)
      }
    }
  }
}

#' w2dist
#'
#' The 2-Wasserstein distance between two multivariate normal distributions
#'
#' @param P A multivariate normal distribution given as a list with arguments mean and cov.
#' @param Q A multivariate normal distribution given as a list with arguments mean and cov.
#'
#' @return A double giving the 2-Wasserstein distance between the two distributions.
#'
#' @examples
#' P <- list(mean = c(1, 1), cov = diag(1, 2))
#' Q <- list(mean = c(0, 0), cov = 1.1*diag(1, 2))
#' Q <- list(mean = c(0, 0), cov = 1.1*diag(1, 2))
#' w2dist(P, Q)
#'
#' @export
#'
w2dist <- function(P, Q) {
    sqrt(abs(optimalFlow:::distGaussianMean(P, Q) + optimalFlow:::distGaussianCov(P, Q)))
}

#' labelTransfer
#'
#' Label transfer between a test partition and a training set of partitions.
#'
#' @param training.cytometry List of partitions, where each partition is a dataframe where the last column contains the labels of the partition.
#' @param test.cytometry Test data, a dataframe without labels.
#' @param test.partition Labels of a partition of the test data.
#' @param equal.weights If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.
#'
#' @return A fuzzy relabeling consistent of a transportation plan.
#'
#' @examples
#' \donttest{
#' data.example <- data.frame(v1 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)),
#'                           v2 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)), id = c(rep(0, 50), rep(1, 50)))
#' test.labels <- c(rep('a', 50), rep('b', 50))
#' labelTransfer(data.example, data.example[, 1:2], test.labels)
#' }
#' @importFrom stats rnorm
#' @export
#'
labelTransfer <- function(training.cytometry, test.cytometry, test.partition, equal.weights = FALSE) {
    dim.cyto <- dim(test.cytometry)[2]
    training.partition <- training.cytometry[, dim(training.cytometry)[2]]
    training.cells <- names(table(training.partition))
    M <- length(training.cells)
    test.cells <- names(table(test.partition))
    N <- length(test.cells)
    cost.matrix <- array(dim = c(M, N))
    t0 <- Sys.time()
    j <- 0
    for (cell in training.cells) {
        j <- j + 1
        n_1 <- sum(training.partition == cell)
        jj <- 0
        for (k_m in test.cells) {
            jj <- jj + 1
            n_2 <- sum(test.partition == k_m)
            if (n_1 <= 10000) {
                if (n_2 <= 10000) {
                  cost <- sum(Rfast::dista(training.cytometry[training.partition == cell, 1:dim.cyto], test.cytometry[test.partition ==
                    k_m, 1:dim.cyto]))/(n_1 * n_2)
                } else {
                  cost <- sum(Rfast::dista(training.cytometry[training.partition == cell, 1:dim.cyto], test.cytometry[sample(which(test.partition ==
                    k_m), 10000), 1:dim.cyto]))/(n_1 * 10000)
                }
            } else {
                if (n_2 <= 10000) {
                  cost <- sum(Rfast::dista(training.cytometry[sample(which(training.partition == cell), 10000), 1:dim.cyto], test.cytometry[test.partition ==
                    k_m, 1:dim.cyto]))/(n_2 * 10000)
                } else {
                  cost <- sum(Rfast::dista(training.cytometry[sample(which(training.partition == cell), 10000), 1:dim.cyto], test.cytometry[sample(which(test.partition ==
                    k_m), 10000), 1:dim.cyto]))/(10000 * 10000)
                }
            }
            cost.matrix[j, jj] <- cost
        }
    }

    if (equal.weights) {
        A <- rep(1/M, M)
        B <- rep(1/N, N)
        names(A) <- training.cells
        names(B) <- test.cells

    } else {
        A <- table(training.partition)/sum(table(training.partition))
        B <- table(test.partition)/sum(table(test.partition))
    }

    optimal.transp.from.A.to.B <- transport::transport(a = A, b = B, costm = cost.matrix)
    optimal.transp.from.A.to.B$to <- names(B)[optimal.transp.from.A.to.B$to]
    optimal.transp.from.A.to.B$from <- names(A)[optimal.transp.from.A.to.B$from]
    return(optimal.transp.from.A.to.B)
}

#' wassersteinKBarycenter
#'
#'  A wrapper for calculating K-barycenters of multivariate normal distributions with
#'   the 2-Wasserstein distance
#'
#' @param i A dummy variable ment for use with apply.
#' @param k Number k of elements in the k-barycenter.
#' @param alpha Level of trimming.
#' @param initialization Type of initialization in c('rnd', 'plus-plus'). 'rnd' makes
#'  the common random initilaization while 'plus-plus' initializes in a similar fashion to k-means++.
#' @param pooled.clusters List of multivariate normals for which the trimmed k-barycenter should be performed.
#'
#' @return A list with entries:
#' \describe{
#'   \item{wasserstein.var}{A double representing the Wasserstein variation.}
#'   \item{wasserstein.k.barycenter}{List with three elements. Variacion_wasser is Waserstein variation.
#'    Baricentro is a list of k elements, each of which is a member of the k-barycenter.
#'    Cluster is the assignation to each barycenter of the original entries.}
#' }
#'
#' @examples
#' normals <- list(list(mean = c(1, 1), cov = diag(2, 2)), list(mean = c(1, 1), cov = diag(1, 2)),
#' list(mean = c(3, 3), cov = diag(1, 2)))
#' wkb <- wassersteinKBarycenter(1, 2, 0, 'rnd', normals)
#' print(wkb$wasserstein.var)
#' print(wkb$wasserstein.k.barycente)
#'
#'@references E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019)
#' optimalFlow: Optimal-transport approach to flow cytometry gating and population matching.
#'  arXiv:1907.08006
#'
#' @noRd
wassersteinKBarycenter <- function(i = 1, k, alpha = 0, initialization = "rnd", pooled.clusters) {
    bar <- tryCatch(optimalFlow::trimmedKBarycenter(k = k, alpha0 = alpha, type.ini = initialization, pooled.clusters), error = function(x) list(bar = "error", 
        variacion_wasser = Inf))
    variation <- bar$variacion_wasser
    return(list(wasserstein.var = variation, wasserstein.k.barycenter = bar))
}

#' optimalFlowClassification
#'
#' Performs a supervised classification of input data when a database and a partition of the database are provided.
#'
#' @param X Datasample to be classified.
#' @param database A list where each entry is a partition (clustering) represented as dataframe, of the same dimensions, where the last variable represents the labels of the partition.
#' @param templates List of the consensus clusterings for every group in the partition of the database obtained by optimalFlowTemplates
#' @param consensus.method The consensus.method value that was used in optimalFlowTemplates.
#' @param cov.estimation How to estimate covariance matrices in each cluster of a partition. "standard" is for using cov(), while "robust" is for using robustbase::covMcd.
#' @param alpha.cov Only when cov.estimation = "robust". Indicates the value of alpha in robustbase::covMcd.
#' @param initial.method Indicates how to obtain a partition of X. Takes values in  c("supervized", "unsupervized"). Supervized uses tclust initilized by templates. Unsupevized usese flowMeans.
#' @param max.clusters The maximum numbers of clusters for flowMeans. Only when initial.method = unsupervized.
#' @param alpha.tclust Level of trimming allowed fo tclust. Only when initial.method = supervized.
#' @param restr.factor.tclust Fixes the restr.fact parameter in tclust. Only when initial.method = supervized.
#' @param classif.method Indicates what type of supervised learning we want to do. Takes values on c("matching", "qda", "random forest").
#' @param qda.bar Only if classif.method = "qda". If True then the appropriate consensus clustering (template, prototype) is used for learning. If False, the closest partition in the appropriate group is used.
#' @param cost.function Only if classif.method = "matching". Indicates the cost function, distance between clusters, to be used for label matching.
#' @param cl.paral Number of cores to be used in parallel procedures.
#' @param equal.weights.voting only when classif.method = "qda" and qda.bar =F, or when  classif.method = "random forest". Indicates the weights structure when looking for the most similar partition in a group.
#' @param equal.weights.template If True, weights assigned to every cluster in a partion are uniform (1/number of clusters). If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.
#'
#' @return A list formed by:
#' \describe{
#' \item{cluster}{Labels assigned to the input data.}
#' \item{clusterings}{A list that contains the initial unsupervized or semi-supervized clusterings of the cytometry of interest. Can have as much entries as the number of templates in the semi-supervized case (initial.method = "supervized), or only one entry in the case of initial.method = "unsupervized". Each entry is a list where the most relevant argument for the clusterings is cluster.}
#' \item{assigned.template.index}{Label of the group for which the template is closer to the data. When classical qda or random forest ares used for classification there is a secon argument indicating the index of the cytometry in the cluster used for learning.}
#' \item{cluster.vote}{Only when classif.method = "matching" or when consensus.method in c("hierarchical", "k-barycenter"). Vote on the type of every label in the partition of the data. In essence, cluster + cluster.vote return a fuzzy clustering of the data of interest.}
#' }
#'
#' @examples
#' # # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
#' database <- buildDatabase(
#'   dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
#'     population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
#' # # To select the appropriate number of templates, via hierarchical tree, in an interactive fashion and produce a clustering we can also use:
#' # templates.optimalFlow <- optimalFlowTemplates(database = database)
#' templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
#'                                              cl.paral = 1)
#' classification.optimalFlow <- optimalFlowClassification(Cytometry1[
#'   which(match(Cytometry1$`Population ID (name)`,c("Monocytes", "CD4+CD8-", "Mature SIg Kappa",
#'                                                   "TCRgd-"), nomatch = 0) > 0), 1:10], database, templates.optimalFlow, cl.paral = 1)
#' scoreF1.optimalFlow <- optimalFlow::f1Score(classification.optimalFlow$cluster,
#'                                            Cytometry1[which(match(Cytometry1$`Population ID (name)`,
#'                                                                                  c("Monocytes", "CD4+CD8-", "Mature SIg Kappa", "TCRgd-"), nomatch = 0) > 0),], noise.types)
#'
#'
#' @references E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
#'
#' @import dplyr
#' @import optimalFlowData
#' @import rlang
#' @importFrom foreach %dopar%
#' @importFrom foreach %:%
#' @importFrom stats as.dist
#' @importFrom stats cutree
#' @importFrom stats cov
#' @importFrom stats hclust
#' @importFrom stats runif
#' @importFrom stats predict
#' @importFrom Rfast dista
#' @importFrom Rfast rowMaxs
#' @importFrom flowMeans flowMeans
#' @export
#'
optimalFlowClassification <- function (X, database, templates, consensus.method = "pooling", cov.estimation = "standard",
                                      alpha.cov = 0.85, initial.method = "supervized", max.clusters = NA, alpha.tclust = 0,
                                      restr.factor.tclust = 1000, classif.method = "qda", qda.bar = TRUE,
                                      cost.function = "points", cl.paral = 1, equal.weights.voting = TRUE,
                                      equal.weights.template = TRUE) {
    t0 <- Sys.time()
    op.syst <- .Platform$OS.type

    ####################### Step 0: Checking that entries are well defined #######################

    if (!is.matrix(X) & !is.data.frame(X)) {
        stop("Data is not a matrix or a dataframe.")
    }
    cov.estimation <- match.arg(cov.estimation, c("standard", "robust"))
    if (!is.double(alpha.cov)) {
        stop("alha.cov is not a doble")
    } else {
        if (alpha.cov < 0 | alpha.cov >= 1) {
            stop("alpha.cov is not in the range (0,1).")
        }
    }
    initial.method <- match.arg(initial.method, c("supervized", "unsupervized"))
    if (!is.double(alpha.tclust)) {
        stop("alha.tclust is not a doble")
    } else {
        if (alpha.tclust < 0 | alpha.tclust >= 1) {
            stop("alpha.tclust is not in the range [0,1).")
        }
    }
    classif.method <- match.arg(classif.method, c("matching", "qda", "random forest"))
    cost.function <- match.arg(cost.function, c("points", "ellipses"))
    op.syst <- match.arg(op.syst, c("unix", "windows"))

    ####################### Step 1: tclust step #######################

    sys.time.step1.0 <- Sys.time()
    dim.cyto <- dim(X)[2]
    if (initial.method == "unsupervized") {
        # stop('unsupervized initialization is not yet implemented.')
        assigned.tclust.index <- 1
        tclust.results <- list()
        tclust.results[[assigned.tclust.index]] <- list(cluster = flowMeans::flowMeans(X, MaxN = max.clusters, nstart = 10, Standardize = FALSE)@Label)

    } else {
        if (length(templates$templates) < cl.paral) {
            n.cores <- length(templates$templates)
        } else {
            n.cores <- cl.paral
        }
        if (op.syst == "unix") {
            tclust.results <- parallel::mclapply(templates$templates, function(template, alpha.tclust, restr.factor.tclust) {
                initial.weights <- unlist(lapply(template, function(x) x$weight))
                initial.means <- matrix(unlist(lapply(template, function(x) x$mean)), ncol = length(initial.weights), byrow = FALSE)

                initial.covs <- array(dim = c(dim.cyto, dim.cyto, length(initial.weights)))
                for (i in 1:length(initial.weights)) {
                  initial.covs[, , i] <- template[[i]]$cov
                }

                template.tclust <- list()
                template.tclust$cov <- initial.covs
                template.tclust$centers <- initial.means
                template.tclust$weights <- initial.weights/sum(initial.weights)
                clustering.tclust <- optimalFlow::tclustWithInitialization(template.tclust, X, i.sol.type = "barycenter", trimming = alpha.tclust,
                  restr.fact = restr.factor.tclust)
                return(clustering.tclust)
            }, alpha.tclust = alpha.tclust, restr.factor.tclust = restr.factor.tclust, mc.cores = n.cores)
        } else {
            cl <- parallel::makePSOCKcluster(n.cores)
            tclust.results <- parallel::parLapply(cl, templates$templates, function(template, alpha.tclust, restr.factor.tclust) {
                initial.weights <- unlist(lapply(template, function(x) x$weight))
                initial.means <- matrix(unlist(lapply(template, function(x) x$mean)), ncol = length(initial.weights), byrow = FALSE)

                initial.covs <- array(dim = c(dim.cyto, dim.cyto, length(initial.weights)))
                for (i in 1:length(initial.weights)) {
                  initial.covs[, , i] <- template[[i]]$cov
                }

                template.tclust <- list()
                template.tclust$cov <- initial.covs
                template.tclust$centers <- initial.means
                template.tclust$weights <- initial.weights/sum(initial.weights)
                clustering.tclust <- optimalFlow::tclustWithInitialization(template.tclust, X, i.sol.type = "barycenter", trimming = alpha.tclust,
                  restr.fact = restr.factor.tclust)
                return(clustering.tclust)
            }, alpha.tclust = alpha.tclust, restr.factor.tclust = restr.factor.tclust)
            parallel::stopCluster(cl)
        }
        objective.tclust <- unlist(lapply(tclust.results, function(x) x$obj))
        assigned.tclust.index <- which.max(objective.tclust)

    }

    sys.time.step1.1 <- Sys.time()
    time.dif.1 <- difftime(sys.time.step1.1, sys.time.step1.0)
    print(paste("step 1:", time.dif.1, units(time.dif.1), sep = " "))

    ####################### Step 2: Similarity distance assignation #######################

    sys.time.step2.0 <- Sys.time()

    data.elliptical <- lapply(sort(unique(tclust.results[[assigned.tclust.index]]$cluster)), optimalFlow::estimCovCellGeneral, cytometry = X,
        labels = tclust.results[[assigned.tclust.index]]$cluster, type = cov.estimation, alpha = alpha.cov)
    data.elliptical <- data.elliptical[!is.na(data.elliptical)]

    similarity.distances <- optimalFlow::costWasserMatchingEllipse(data.elliptical, templates$templates, equal.weights.template)
    print("Similarity distances to templates:")
    print(similarity.distances)
    assigned.template.index <- which.min(similarity.distances)
    assigned.template <- templates$templates[[assigned.template.index]]

    sys.time.step2.1 <- Sys.time()
    time.dif.2 <- difftime(sys.time.step2.1, sys.time.step2.0)
    print(paste("step 2:", time.dif.2, units(time.dif.2), sep = " "))

    ####################### Step 3: Labelling #######################

    sys.time.step3.0 <- Sys.time()

    if (classif.method == "qda" & qda.bar) {
        weight.groups <- unlist(lapply(assigned.template, function(x) x$weight))
        if (abs(sum(weight.groups) - 1) > 10^(-8)) {
            for (i in 1:length(assigned.template)) {
                assigned.template[[i]]$weight <- assigned.template[[i]]$weight/sum(weight.groups)
            }
        }
        if (length(assigned.template) < cl.paral) {
            n.cores <- length(assigned.template)
        } else {
            n.cores <- cl.paral
        }
        if (op.syst == "unix") {
            qda.assignation <- parallel::mclapply(assigned.template, optimalFlow::qdaClassification, data = X, mc.cores = n.cores)
        } else {
            cl <- parallel::makePSOCKcluster(n.cores)
            qda.assignation <- parallel::parLapply(cl, assigned.template, optimalFlow::qdaClassification, data = X)
            parallel::stopCluster(cl)
        }
        qda.assignation <- qda.assignation[!is.na(qda.assignation)]
        qda.assignation <- matrix(unlist(qda.assignation), ncol = length(qda.assignation), byrow = FALSE)
        qda.groups <- Rfast::rowMaxs(qda.assignation)
        unique.qda.groups <- sort(unique(qda.groups))
        qda.groups <- factor(qda.groups)
        levels(qda.groups) <- unlist(lapply(assigned.template, function(x) x$type))[!is.na(qda.assignation)]

        if (consensus.method == "hierarchical" | consensus.method == "k-barycenter") {
            vote <- list()
            if (sum(is.na(as.integer(levels(qda.groups)))) > 0) {
                range.qda.groups <- levels(qda.groups)
            } else {
                range.qda.groups <- as.integer(levels(qda.groups))
            }
            for (i in range.qda.groups) {
                vote[[as.character(i)]] <- data.frame(cell = names(assigned.template[[which(range.qda.groups == i)]]$type.proportions),
                  simple.proportion = as.vector(assigned.template[[which(range.qda.groups == i)]]$type.proportions))
            }
            sys.time.step3.1 <- Sys.time()
            time.dif.3 <- difftime(sys.time.step3.1, sys.time.step3.0)
            print(paste("step 3:", time.dif.3, units(time.dif.3), sep = " "))
            t1 <- Sys.time()
            print(difftime(t1, t0))
            return(list(cluster = qda.groups, cluster.vote = vote, vote.original.proportion = lapply(assigned.template, function(x) x$type.proportions),
                clusterings = tclust.results, assigned.template.index = assigned.template.index))
        } else {
            sys.time.step3.1 <- Sys.time()
            time.dif.3 <- difftime(sys.time.step3.1, sys.time.step3.0)
            print(paste("step 3:", time.dif.3, units(time.dif.3), sep = " "))
            t1 <- Sys.time()
            print(difftime(t1, t0))
            return(list(cluster = qda.groups, clusterings = tclust.results, assigned.template.index = assigned.template.index))
        }
    } else {
        if (classif.method == "matching") {
            if (cost.function == "points") {
                cytos.cluster <- templates$clustering
                if (sum(cytos.cluster == assigned.template.index) < cl.paral) {
                  n.cores <- sum(cytos.cluster == assigned.template.index)
                } else {
                  n.cores <- cl.paral
                }
                vote <- optimalFlow::voteLabelTransfer(type = cost.function, test.partition = tclust.results[[assigned.tclust.index]]$cluster,
                  test.cytometry = X, training.cytometries = database[cytos.cluster == assigned.template.index], test = 1, op.syst = op.syst,
                  cl.paral = n.cores, equal.weights = equal.weights.voting)
                sys.time.step3.1 <- Sys.time()
                time.dif.3 <- difftime(sys.time.step3.1, sys.time.step3.0)
                print(paste("step 3:", time.dif.3, units(time.dif.3), sep = " "))
                t1 <- Sys.time()
                print(difftime(t1, t0))
                return(list(cluster = tclust.results[[assigned.tclust.index]]$cluster, cluster.vote = vote$final.vote[[1]], clusterings = tclust.results,
                  assigned.template.index = assigned.template.index))

            } else {
                cytos.cluster <- templates$clustering
                vote <- optimalFlow::voteLabelTransfer(type = cost.function, test.partition.ellipse = data.elliptical, training.cytometries.barycenter = assigned.template,
                  test = 1, op.syst = op.syst, cl.paral = 1, equal.weights = equal.weights.voting)
                sys.time.step3.1 <- Sys.time()
                time.dif.3 <- difftime(sys.time.step3.1, sys.time.step3.0)
                print(paste("step 3:", time.dif.3, units(time.dif.3), sep = " "))
                t1 <- Sys.time()
                print(difftime(t1, t0))
                if (consensus.method == "hierarchical" | consensus.method == "k-barycenter") {
                  vote.original.proportion <- lapply(assigned.template, function(x) x$type.proportions)
                  if (sum(cytos.cluster == assigned.template.index) == 1) {
                    # print(vote) print(vote$final.vote[[1]]) final.vote = vote$fianl.vote[[1]] print(final.vote)
                    return(list(cluster = tclust.results[[assigned.tclust.index]]$cluster, cluster.vote = vote$final.vote[[1]],
                      clusterings = tclust.results, assigned.template.index = assigned.template.index))
                  } else {
                    final.vote <- voteTransformation(vote$final.vote[[1]], vote.original.proportion)
                    return(list(cluster = tclust.results[[assigned.tclust.index]]$cluster, cluster.vote = final.vote, clusterings = tclust.results,
                      assigned.template.index = assigned.template.index))
                  }
                } else {
                  return(list(cluster = tclust.results[[assigned.tclust.index]]$cluster, cluster.vote = vote$final.vote[[1]], clusterings = tclust.results,
                    assigned.template.index = assigned.template.index))
                }
            }
        } else {
            cytos.cluster <- templates$clustering
            database.elliptical <- templates$database.elliptical
            similarity.distances <- optimalFlow::costWasserMatchingEllipse(data.elliptical, database.elliptical[cytos.cluster ==
                assigned.template.index], equal.weights.template)
            assigned.training.cytometry.index <- which.min(similarity.distances)
            if (classif.method == "qda") {
                assigned.training.cytometry <- database.elliptical[cytos.cluster == assigned.template.index][[assigned.training.cytometry.index]]
                if (length(assigned.training.cytometry) < cl.paral) {
                  n.cores <- length(assigned.training.cytometry)
                } else {
                  n.cores <- cl.paral
                }
                if (op.syst == "unix") {
                  qda.assignation <- parallel::mclapply(assigned.training.cytometry, optimalFlow::qdaClassification, data = X, mc.cores = n.cores)
                } else {
                  cl <- parallel::makePSOCKcluster(n.cores)
                  qda.assignation <- parallel::parLapply(cl, assigned.training.cytometry, optimalFlow::qdaClassification, data = X)
                  parallel::stopCluster(cl)
                }
                qda.assignation <- qda.assignation[!is.na(qda.assignation)]
                qda.assignation <- matrix(unlist(qda.assignation), ncol = length(qda.assignation), byrow = FALSE)
                qda.groups <- Rfast::rowMaxs(qda.assignation)
                unique.qda.groups <- sort(unique(qda.groups))
                qda.groups <- factor(qda.groups)
                levels(qda.groups) <- unlist(lapply(assigned.training.cytometry, function(x) x$type))[!is.na(qda.assignation)]

                sys.time.step3.1 <- Sys.time()
                time.dif.3 <- difftime(sys.time.step3.1, sys.time.step3.0)
                print(paste("step 3:", time.dif.3, units(time.dif.3), sep = " "))
                t1 <- Sys.time()
                print(difftime(t1, t0))
                return(list(cluster = qda.groups, clusterings = tclust.results, assigned.template.index = c(assigned.template.index,
                  assigned.training.cytometry.index)))
            } else {
                assigned.training.cytometry <- database[cytos.cluster == assigned.template.index][[assigned.training.cytometry.index]]
                # ytest = as.factor(citometria$`Population ID (name)`)
                dim.cyto <- dim(assigned.training.cytometry)[2]
                y <- droplevels(as.factor(dplyr::pull(assigned.training.cytometry, dim.cyto)))
                training.rforest <- randomForest::randomForest(assigned.training.cytometry[, 1:(dim.cyto - 1)], y = y, keep.forest = TRUE)
                random.forest.groups <- predict(training.rforest, X)
                sys.time.step3.1 <- Sys.time()
                time.dif.3 <- difftime(sys.time.step3.1, sys.time.step3.0)
                print(paste("step 3:", time.dif.3, units(time.dif.3), sep = " "))
                t1 <- Sys.time()
                print(difftime(t1, t0))
                return(list(cluster = random.forest.groups, clusterings = tclust.results, assigned.template.index = c(assigned.template.index,
                  assigned.training.cytometry.index)))
            }
        }
    }
}

#' optimalFlowTemplates
#'
#' Returns a partition of the input clusterings with a respective consensus clustering for every group.
#'
#' @param database A list where each entry is a partition (clustering) represented as dataframe, of the same dimensions, where the last variable represents the labels of the partition.
#' @param database.names Names of the elements in the database.
#' @param cov.estimation How to estimate covariance matrices in each cluster of a partition. 'standard' is for using cov(), while 'robust' is for using robustbase::covMcd.
#' @param alpha.cov Only when cov.estimation = 'robust'. Indicates the value of alpha in robustbase::covMcd.
#' @param equal.weights.template If True, weights assigned to every cluster in a partion are uniform (1/number of clusters). If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.
#' @param hclust.method Indicates what kind of hierarchical clustering to  do with the similarity distances matrix of the partitions. Takes values in c('complete', 'single', 'average', 'hdbscan', 'dbscan').
#' @param trimm.template Logical value. Indicates if it is allowed to not take into account some of the entries of database. Default is False.
#' @param templates.number Only if hclust.method in c('complete', 'single', 'average'). Indicates the number of clusters to use with cutree. If set to NA (default), plots the hierarchical tree and asks the user to introduce an appropriate number of clusters.
#' @param minPts Only if hclust.method in c('hdbscan', 'dbscan'). Indicates the value of argument minPts in dbscan::dbscan and dbscan::hdbscan.
#' @param eps Only if hclust.method = 'dbscan'. Indicates the value of eps in dbscan::dbscan.
#' @param consensus.method Sets the way of doing consensus clustering when clusters are viewed as Multivariate Distributions. Can take values in c('pooling', 'k-barycenter', 'hierarchical'). See details.
#' @param barycenters.number Only if consensus.method = 'k-barycenter'. Sets the number, k, of barycenters when using k-barycenters.
#' @param bar.repetitions Only if consensus.method = 'k-barycenter'. How many times to repeat the k-barycenters procedure. Equivalent to nstart in kmeans.
#' @param alpha.bar Only if consensus.method = 'k-barycenter'. The level of trimming allowed during the k-barycenters procedure.
#' @param bar.ini.method Only if consensus.method = 'k-barycenter'. Takes values in c('rnd', 'plus-plus'). See details.
#' @param consensus.minPts Only if consensus.method = 'hierarchical'. The value of argument minPts for dbscan::hdbscan.
#' @param cl.paral Number of cores to be used in parallel procedures.
#'
#' @return A list containting:
#' \describe{
#'  \item{templates}{A list representing the consensus clusterings for every group in the partition of the database. Each element of the list is a template partition. Hence it is a list itself, containig the cell types in the prototype, where each element has components: mean, cov, weight and type.}
#'  \item{clustering}{Clustering of the input partitions.}
#'  \item{database.elliptical}{A list containig each cytometry in the database viewed as a mixture distribution. Each element of the list is a cytometry viewed as a mixture. Hence it is a list itself, containig the cell types in the cytometry, where each element has components: mean, cov, weight and type.}
#' }
#'
#' @examples
#' # # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
#' database <- buildDatabase(
#'   dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
#'     population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
#'
#' # # To select the appropriate number of templates, via hierarchical tree, in an interactive fashion and produce a clustering we can also use:
#' # templates.optimalFlow <- optimalFlowTemplates(database = database)
#'
#' templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
#'                                              cl.paral = 1)
#'
#' @references E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
#'
#' @import dplyr
#' @import optimalFlowData
#' @import rlang
#' @importFrom foreach %dopar%
#' @importFrom foreach %:%
#' @importFrom stats as.dist
#' @importFrom stats cutree
#' @importFrom stats cov
#' @importFrom stats hclust
#' @importFrom stats runif
#'
#' @export
#'
optimalFlowTemplates <- function(database, database.names = NULL, cov.estimation = "standard", alpha.cov = 0.85, equal.weights.template = TRUE,
    hclust.method = "complete", trimm.template = FALSE, templates.number = NA, minPts = 2, eps = 1, consensus.method = "pooling",
    barycenters.number = NA, bar.repetitions = 40, alpha.bar = 0.05, bar.ini.method = "plus-plus", consensus.minPts = 3, cl.paral = 1) {
    t0 <- Sys.time()
    op.syst <- .Platform$OS.type

    ####################### Step 0: Checking that entries are well defined #######################

    cov.estimation <- match.arg(cov.estimation, c("standard", "robust"))
    if (!is.double(alpha.cov)) {
        stop("alha.cov is not a doble")
    } else {
        if (alpha.cov < 0 | alpha.cov >= 1) {
            stop("alpha.cov is not in the range (0,1).")
        }
    }
    hclust.method <- match.arg(hclust.method, c("complete", "single", "average", "hdbscan", "dbscan"))
    if (!is.na(templates.number)) {
        if (!is.double(templates.number) & !is.integer(templates.number)) {
            stop("templates.number is not well defined.")
        } else {
            if (templates.number < 1) {
                stop("templates.number should be >= 1.")
            }
        }
    }
    if (hclust.method %in% c("hdbscan", "dbscan")) {
        minPts <- as.integer(minPts)
        if (is.na(minPts) | is.na(eps)) {
            stop("minPts or eps not well defined.")
        }
    }
    consensus.method <- match.arg(consensus.method, c("pooling", "k-barycenter", "hierarchical"))
    if (consensus.method == "k-barycenter") {
        barycenters.number <- as.integer(barycenters.number)
        bar.repetitions <- as.integer(bar.repetitions)
        if (is.na(bar.repetitions)) {
            stop("bar.repetitions must be an integer >= 1.")
        }
        if (!is.double(alpha.bar)) {
            stop("alha.bar is not a doble")
        } else {
            if (alpha.bar < 0 | alpha.bar >= 1) {
                stop("alpha.bar is not in the range [0,1).")
            }
        }
        bar.ini.method <- match.arg(bar.ini.method, c("rnd", "plus-plus"))
    } else {
        if (consensus.method == "hierarchical") {
            consensus.minPts <- as.integer(consensus.minPts)
            if (is.na(consensus.minPts)) {
                stop("consensus.minPts or consensus.eps not well defined.")
            }
        }
    }
    op.syst <- match.arg(op.syst, c("unix", "windows"))

    ####################### Step 1: Similarity distance based clustering #######################

    sys.time.step1.0 <- Sys.time()
    database.elliptical <- list()
    for (i in 1:length(database)) {
        dim.cyto <- dim(database[[i]])[2]
        database.elliptical[[i]] <- lapply(names(table(database[[i]][, dim.cyto])), estimCovCellGeneral, cytometry = database[[i]][,
            1:(dim.cyto - 1)], labels = database[[i]][, dim.cyto], type = cov.estimation, alpha = alpha.cov)
        database.elliptical[[i]] <- database.elliptical[[i]][!is.na(database.elliptical[[i]])]
    }

    # cl <- parallel::makePSOCKcluster(cl.paral) doParallel::registerDoParallel(cl) database.elliptical = foreach::foreach(i =
    # 1:length(database)) %dopar% { source('estimCovCellGeneral.R') database.elliptical0 = lapply(names(table(database[[i]][,
    # dim.cyto])), estimCovCellGeneral, cytometry = database[[i]][, 1:(dim.cyto - 1)], labels = database[[i]][, dim.cyto], type =
    # cov.estimation, alpha = alpha.cov) database.elliptical0 = database.elliptical0[!is.na(database.elliptical0)] }
    # parallel::stopCluster(cl)

    dim.cytos <- length(database.elliptical)
    if (!is.na(templates.number) & dim.cytos < templates.number){
      stop(paste("templates.number, currently ", templates.number,", should be betwen 1 and ", dim.cytos," (the number of elements in the database).", sep = ""))
    }
    wasser.cost <- array(0, dim = c(dim.cytos, dim.cytos))
    cytometries_elipses <- list()
    if ((dim.cytos - 2) * (dim.cytos - 2 + 1)/2 < cl.paral) {
        n.cores <- (dim.cytos - 2) * (dim.cytos - 2 + 1)/2
    } else {
        n.cores <- cl.paral
    }
    cl <- parallel::makePSOCKcluster(n.cores, setup_strategy = "sequential")
    doParallel::registerDoParallel(cl)
    transport_costs_list <- foreach::foreach(j = 1:(dim.cytos - 1)) %:% foreach::foreach(i = (j + 1):dim.cytos) %dopar% {
        optimalFlow::wasserCostFunction(j, i, database.elliptical, equal.weights.template)
    }
    wasser.cost[lower.tri(wasser.cost, diag = FALSE)] <- unlist(transport_costs_list)
    parallel::stopCluster(cl)


    if (hclust.method %in% c("complete", "single", "average")) {
        cytos.hcl <- hclust(as.dist(wasser.cost), method = hclust.method)
        if (is.na(templates.number)) {
            if (is.null(database.names)) {
                graphics::plot(cytos.hcl, main = NULL, ylab = NULL, xlab = "")
            } else {
                graphics::plot(cytos.hcl, labels = database.names, main = NULL, ylab = NULL, xlab = "")
            }
            templates.number <- as.integer(readline(prompt = "How many clusters should I look for : "))
            if (!is.na(templates.number) & dim.cytos < templates.number){
              stop(paste("templates.number, currently ", templates.number,", should be betwen 1 and ", dim.cytos," (the number of elements in the database).", sep = ""))
            }
            cytos.cluster <- cutree(cytos.hcl, k = templates.number)
        } else {
            cytos.cluster <- cutree(cytos.hcl, k = templates.number)
        }
    } else {
        if (hclust.method == "hdbscan") {
            cytos.cluster <- dbscan::hdbscan(as.dist(wasser.cost), minPts = minPts)$cluster
        } else {
            cytos.cluster <- dbscan::dbscan(as.dist(wasser.cost), minPts = minPts, eps = eps)$cluster
        }
    }
    if (!trimm.template) {
        indexes.trimmed <- which(cytos.cluster == 0)
        if (length(indexes.trimmed) > 0) {
            cytos.cluster <- cytos.cluster + length(indexes.trimmed)
            cytos.cluster[indexes.trimmed] <- seq(1, length(indexes.trimmed), 1)
        }
    }
    sys.time.step1.1 <- Sys.time()
    time.dif.1 <- difftime(sys.time.step1.1, sys.time.step1.0)
    print(paste("step 1:", time.dif.1, units(time.dif.1), sep = " "))

    ####################### Step 2: Consensus clustering #######################

    sys.time.step2.0 <- Sys.time()
    templates <- list()
    templates.cell.types <- list()
    if (consensus.method == "pooling") {
        ########## pooling clusters with the same labels and taking barycenter #########
        j <- 0
        for (i in sort(unique(cytos.cluster))) {
            if (i == 0) {
                next
            } else {
                j <- j + 1
                cell.types <- lapply(database.elliptical[cytos.cluster == i], function(x) lapply(x, function(y) y$type))
                cell.types <- sort(unique(unlist(cell.types)))
                templates.cell.types[[j]] <- cell.types
                if (length(cell.types) < cl.paral) {
                  n.cores <- length(cell.types)
                } else {
                  n.cores <- cl.paral
                }
                if (op.syst == "unix") {
                  templates[[j]] <- parallel::mclapply(cell.types, optimalFlow::estimationCellBarycenter, database.elliptical[cytos.cluster ==
                    i], mc.cores = n.cores)
                } else {
                  cl <- parallel::makePSOCKcluster(n.cores)
                  templates[[j]] <- parallel::parLapply(cl, cell.types, optimalFlow::estimationCellBarycenter, database.elliptical[cytos.cluster ==
                    i])
                  parallel::stopCluster(cl)
                }
            }
        }
    } else {
        ######## pooling all clusters for the partitions in the same group #######
        pooled.clustered.cytometries <- list()
        for (i in sort(unique(cytos.cluster))) {
            pooled.clustered.cytometries.0 <- list()
            l <- 0
            for (training.cytometry in database.elliptical[cytos.cluster == i]) {
                for (k in 1:length(training.cytometry)) {
                  l <- l + 1
                  pooled.clustered.cytometries.0[[l]] <- training.cytometry[[k]]
                }
            }
            pooled.clustered.cytometries[[as.character(i)]] <- pooled.clustered.cytometries.0
        }
        if (consensus.method == "hierarchical") {
            ######## Density based hierarchical clustering based on the wasserstein distance between clusters. For each resulting grouping we take
            ######## the one-barycenter.
            j <- 0
            for (i in sort(unique(cytos.cluster))) {
                if (i == 0) {
                  next
                } else {
                  j <- j + 1
                  if (sum(cytos.cluster == i) == 1) {
                    cell.types <- lapply(database.elliptical[cytos.cluster == i], function(x) lapply(x, function(y) y$type))
                    cell.types <- sort(unique(unlist(cell.types)))
                    template.partition <- cell.types
                  } else {
                    dim.cytos <- length(pooled.clustered.cytometries[[as.character(i)]])
                    wasser.cost.list <- wasser.cost <- array(0, dim = c(dim.cytos, dim.cytos))
                    if ((dim.cytos - 2) * (dim.cytos - 2 + 1)/2 < cl.paral) {
                      n.cores <- (dim.cytos - 2) * (dim.cytos - 2 + 1)/2
                    } else {
                      n.cores <- cl.paral
                    }
                    cl <- parallel::makePSOCKcluster(n.cores, setup_strategy = "sequential")
                    doParallel::registerDoParallel(cl)
                    wasser.cost.list <- foreach::foreach(k = 1:(dim.cytos - 1)) %:% foreach::foreach(l = (k + 1):dim.cytos) %dopar%
                      {
                        optimalFlow::w2dist(pooled.clustered.cytometries[[as.character(i)]][[k]], pooled.clustered.cytometries[[as.character(i)]][[l]])
                      }
                    wasser.cost[lower.tri(wasser.cost, diag = FALSE)] <- unlist(wasser.cost.list)
                    parallel::stopCluster(cl)
                    template.partition <- dbscan::hdbscan(as.dist(wasser.cost), minPts = consensus.minPts)$cluster
                  }
                  template.partition.labels <- sort(unique(template.partition))
                  if (length(which(template.partition.labels == 0)) > 0) {
                    template.partition.labels <- template.partition.labels[-1]
                  }
                  cl <- parallel::makePSOCKcluster(cl.paral, setup_strategy = "sequential")
                  doParallel::registerDoParallel(cl)
                  template <- foreach::foreach(k = template.partition.labels) %dopar% {
                    pooled.elements <- pooled.clustered.cytometries[[as.character(i)]][template.partition == k]
                    pooled.elements.type.proportion <- unlist(lapply(pooled.elements, function(x) x$type))
                    pooled.elements.type.proportion <- table(pooled.elements.type.proportion)
                    pooled.elements.type.proportion <- sort(pooled.elements.type.proportion/sum(pooled.elements.type.proportion),
                      decreasing = TRUE)
                    pooled.elements.covs <- lapply(pooled.elements, function(x) x$cov)
                    pooled.elements.weighted.means <- matrix(unlist(lapply(pooled.elements, function(x) x$mean)), ncol = length(pooled.elements),
                      byrow = FALSE)
                    pooled.elements.weights <- unlist(lapply(pooled.elements, function(x) x$weight))
                    mean.weight <- mean(pooled.elements.weights)
                    pooled.elements.weights <- pooled.elements.weights/sum(pooled.elements.weights)
                    pooled.mean <- colSums(t(pooled.elements.weighted.means) * pooled.elements.weights)
                    template.cell <- list(mean = pooled.mean, cov = optimalFlow:::GaussianBarycenters(pooled.elements.covs, pooled.elements.weights)$Barycenter,
                      weight = mean.weight, type = k, type.proportions = pooled.elements.type.proportion)
                  }
                  parallel::stopCluster(cl)
                  templates[[j]] <- template
                  templates.cell.types[[j]] <- template.partition.labels
                }
            }
        } else {
            j <- 0
            for (i in sort(unique(cytos.cluster))) {
                if (i == 0) {
                  next
                } else {
                  j <- j + 1
                  if (sum(cytos.cluster == i) == 1) {
                    cell.types <- lapply(database.elliptical[cytos.cluster == i], function(x) lapply(x, function(y) y$type))
                    cell.types <- sort(unique(unlist(cell.types)))
                    template.partition <- cell.types
                    template.partition.labels <- sort(unique(template.partition))
                    if (length(which(template.partition.labels == 0)) > 0) {
                      template.partition.labels <- template.partition.labels[-1]
                    }
                    cl <- parallel::makePSOCKcluster(cl.paral, setup_strategy = "sequential")
                    doParallel::registerDoParallel(cl)
                    template <- foreach::foreach(k = template.partition.labels) %dopar% {
                      pooled.elements <- pooled.clustered.cytometries[[as.character(i)]][template.partition == k]
                      pooled.elements.type.proportion <- unlist(lapply(pooled.elements, function(x) x$type))
                      pooled.elements.type.proportion <- table(pooled.elements.type.proportion)
                      pooled.elements.type.proportion <- sort(pooled.elements.type.proportion/sum(pooled.elements.type.proportion),
                        decreasing = TRUE)
                      pooled.elements.covs <- lapply(pooled.elements, function(x) x$cov)
                      pooled.elements.weighted.means <- matrix(unlist(lapply(pooled.elements, function(x) x$mean)), ncol = length(pooled.elements),
                        byrow = FALSE)
                      pooled.elements.weights <- unlist(lapply(pooled.elements, function(x) x$weight))
                      mean.weight <- mean(pooled.elements.weights)
                      pooled.elements.weights <- pooled.elements.weights/sum(pooled.elements.weights)
                      pooled.mean <- colSums(t(pooled.elements.weighted.means) * pooled.elements.weights)
                      template.cell <- list(mean = pooled.mean, cov = optimalFlow:::GaussianBarycenters(pooled.elements.covs, pooled.elements.weights)$Barycenter,
                        weight = mean.weight, type = k, type.proportions = pooled.elements.type.proportion)
                    }
                    parallel::stopCluster(cl)
                    templates[[j]] <- template
                    templates.cell.types[[j]] <- template.partition.labels
                  } else {
                    if (bar.repetitions < cl.paral) {
                      n.cores <- bar.repetitions
                    } else {
                      n.cores <- cl.paral
                    }
                    if (op.syst == "unix") {
                      if (is.na(barycenters.number)) {
                        barycenters.number.0 <- ceiling(mean(unlist(lapply(which(cytos.cluster == i), function(ii) length(database.elliptical[[ii]])))))
                        barycenters.number.0 <- round((1 + 0.1) * barycenters.number.0)
                        k.barycenter.list <- parallel::mclapply(1:bar.repetitions, optimalFlow:::wassersteinKBarycenter, k = barycenters.number.0,
                          alpha = alpha.bar, initialization = "plus-plus", pooled.clustered.cytometries[[as.character(i)]], mc.cores = n.cores)  #n.cores
                      } else {
                        if (barycenters.number > length(pooled.clustered.cytometries[[as.character(i)]])) {
                          stop(paste("barycenters.number, currently ", barycenters.number, ", should be between 1 and ", length(pooled.clustered.cytometries[[as.character(i)]]), " (the total number of elements in the group).", sep = ""))
                        }
                        k.barycenter.list <- parallel::mclapply(1:bar.repetitions, optimalFlow:::wassersteinKBarycenter, k = barycenters.number,
                          alpha = alpha.bar, initialization = "plus-plus", pooled.clustered.cytometries[[as.character(i)]], mc.cores = n.cores)  #n.cores
                      }
                    } else {
                      if (is.na(barycenters.number)) {
                        barycenters.number.0 <- ceiling(mean(unlist(lapply(which(cytos.cluster == i), function(ii) length(database.elliptical[[ii]])))))
                        barycenters.number.0 <- round((1 + 0.1) * barycenters.number.0)
                        cl <- parallel::makePSOCKcluster(n.cores, setup_strategy = "sequential")
                        k.barycenter.list <- parallel::parLapply(cl, 1:bar.repetitions, optimalFlow:::wassersteinKBarycenter, k = barycenters.number.0,
                          alpha = alpha.bar, initialization = "plus-plus", pooled.clustered.cytometries[[as.character(i)]])
                        parallel::stopCluster(cl)
                      } else {
                        if (barycenters.number > length(pooled.clustered.cytometries[[as.character(i)]])) {
                          stop(paste("barycenters.number, currently ", barycenters.number, ", should be between 1 and ", length(pooled.clustered.cytometries[[as.character(i)]]), " (the total number of elements in the group).", sep = ""))
                        }
                        cl <- parallel::makePSOCKcluster(n.cores, setup_strategy = "sequential")
                        k.barycenter.list <- parallel::parLapply(cl, 1:bar.repetitions, optimalFlow:::wassersteinKBarycenter, k = barycenters.number,
                          alpha = alpha.bar, initialization = "plus-plus", pooled.clustered.cytometries[[as.character(i)]])
                        parallel::stopCluster(cl)
                      }
                    }
                    wasser.var <- lapply(1:bar.repetitions, function(ii) k.barycenter.list[[ii]]$wasserstein.var)
                    if (min(unlist(wasser.var)) == Inf) {
                      stop("All executions of k-barycenter were unsuccesfull")
                    } else {
                      k.barycenter <- k.barycenter.list[[which.min(wasser.var)]]$wasserstein.k.barycenter
                      template.partition <- k.barycenter$cluster
                      template.partition.labels <- sort(unique(template.partition))
                      if (length(which(template.partition.labels == 0)) > 0) {
                        template.partition.labels <- template.partition.labels[-1]
                      }

                      if (length(template.partition.labels) < cl.paral) {
                        n.cores <- length(template.partition.labels)
                      } else {
                        n.cores <- cl.paral
                      }
                      if (op.syst == "unix") {
                        template <- parallel::mclapply(template.partition.labels, function(k) {
                          pooled.elements <- pooled.clustered.cytometries[[as.character(i)]][template.partition == k]
                          pooled.elements.type.proportion <- unlist(lapply(pooled.elements, function(x) x$type))
                          pooled.elements.type.proportion <- table(pooled.elements.type.proportion)
                          pooled.elements.type.proportion <- sort(pooled.elements.type.proportion/sum(pooled.elements.type.proportion),
                            decreasing = TRUE)
                          pooled.elements.weights <- unlist(lapply(pooled.elements, function(x) x$weight))
                          mean.weight <- mean(pooled.elements.weights)
                          template.cell <- list(mean = k.barycenter$baricentro[[which(template.partition.labels == k)]]$mean, cov = k.barycenter$baricentro[[which(template.partition.labels ==
                            k)]]$cov, weight = mean.weight, type = k, type.proportions = pooled.elements.type.proportion)
                        }, mc.cores = n.cores)
                      } else {
                        cl <- parallel::makePSOCKcluster(n.cores)
                        template <- parallel::parLapply(cl, template.partition.labels, function(k) {
                          pooled.elements <- pooled.clustered.cytometries[[as.character(i)]][template.partition == k]
                          pooled.elements.type.proportion <- unlist(lapply(pooled.elements, function(x) x$type))
                          pooled.elements.type.proportion <- table(pooled.elements.type.proportion)
                          pooled.elements.type.proportion <- sort(pooled.elements.type.proportion/sum(pooled.elements.type.proportion),
                            decreasing = TRUE)
                          pooled.elements.weights <- unlist(lapply(pooled.elements, function(x) x$weight))
                          mean.weight <- mean(pooled.elements.weights)
                          template.cell <- list(mean = k.barycenter$baricentro[[which(template.partition.labels == k)]]$mean, cov = k.barycenter$baricentro[[which(template.partition.labels ==
                            k)]]$cov, weight = mean.weight, type = k, type.proportions = pooled.elements.type.proportion)
                        })
                        parallel::stopCluster(cl)
                      }
                      templates[[j]] <- template
                      templates.cell.types[[j]] <- template.partition.labels
                    }
                  }
                }
            }
        }
    }
    sys.time.step2.1 <- Sys.time()
    time.dif.2 <- difftime(sys.time.step2.1, sys.time.step2.0)
    print(paste("step 2:", time.dif.2, units(time.dif.2), sep = " "))
    t1 <- Sys.time()
    time.dif.total <- difftime(sys.time.step2.1, t0)
    print(paste("Execution time:", time.dif.total, units(time.dif.total), sep = " "))
    return(list(templates = templates, clustering = cytos.cluster, database.elliptical = database.elliptical))
}

#' tclustWithInitialization
#'
#' A wrapper for the function tclust_H.
#'
#' @param initialization Initial solution for parameters provided by the user. Can be a matrix of data containing observations anc cluster assignations or can be a list spesifying a multivariate mixture of gaussians.
#' @param cytometry A matrix or data.frame of dimension n x p, containing the observations (row-wise).
#' @param i.sol.type Type of initial solutions in c('points', 'barycenters'). 'points' refers to a classified data matrix, while 'barycenters' to a multivariate mixture.
#' @param trimming The proportion of observations to be trimmed.
#' @param restr.fact The constant restr.fact >= 1 constrains the allowed differences among group scatters. Larger values imply larger differences of group scatters, a value of 1 specifies the strongest restriction.
#'
#' @return A list with entries:
#' \describe{
#'  \item{cluster}{A numerical vector of size n containing the cluster assignment for each observation. Cluster names are integer numbers from 1 to k, 0 indicates trimmed observations.}
#'  \item{n_clus}{Number of clusters actually found.}
#'  \item{obj}{he value of the objective function of the best (returned) solution.}
#' }
#'
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2),
#'         matrix(rnorm(100) + 4, ncol = 2))
#' ## robust cluster obtention from a sample x asking for 3 clusters,
#' ## trimming level 0.05 and constrain level 12
#' k <- 3; alpha <- 0.05; restr.fact <- 12
#' output = tclust_H(x = x, k = k, alpha = alpha, nstart = 50, iter.max = 20,
#'                  restr = 'eigen', restr.fact = restr.fact, sol_ini_p = FALSE, sol_ini = NA,
#'                  equal.weights = FALSE, trace = 0, zero.tol = 1e-16)
#' ## cluster assigment
#' output2 <- tclustWithInitialization(data.frame(x, output$cluster), x, 'points', 0.05, 10)
#'
#' @export
#'
tclustWithInitialization <- function(initialization, cytometry, i.sol.type = "points", trimming = 0.05, restr.fact = 1000) {
    t0 <- Sys.time()
    dim.cyto <- dim(cytometry)[2]
    if (i.sol.type == "points") {
        cosa <- data.frame(initialization[, 1:dim.cyto], cluster = dplyr::pull(initialization, (dim.cyto + 1)))
        n_clus <- length(levels(cosa$cluster))
        levels(cosa$cluster) <- 1:n_clus

        cosa_no_noise <- cosa[which(as.double(cosa$cluster) > 0), ]

        sol_inicial_0 <- lapply(1:n_clus, estimCovCellGeneral, cosa[, 1:dim.cyto], cosa$cluster)
        not_valid_estim <- which(is.na(sol_inicial_0))
        if (length(not_valid_estim) > 0) {
            sol_inicial_0 <- sol_inicial_0[-not_valid_estim]
        } else {
            sol_inicial_0 <- sol_inicial_0
        }
        weight_citos <- unlist(lapply(sol_inicial_0, function(x) x$weight))
        means_citos <- matrix(unlist(lapply(sol_inicial_0, function(x) x$mean)), ncol = length(weight_citos), byrow = FALSE)
        cov_citos <- array(dim = c(dim.cyto, dim.cyto, length(weight_citos)))
        for (i in 1:length(weight_citos)) {
            cov_citos[, , i] <- sol_inicial_0[[i]]$cov
        }

        sol_inicial <- list()
        sol_inicial$weights <- weight_citos
        sol_inicial$cov <- cov_citos
        sol_inicial$centers <- means_citos
        KK <- length(sol_inicial$weights)
        print(paste("tclust looking for k = ", KK, sep = ""))
        trimming <- 0.5 * exp(-(KK - 1)/5)
    } else {
        if (i.sol.type == "barycenter") {
            sol_inicial <- initialization
            KK <- length(sol_inicial$weights)
            print(paste("tclust looking for k = ", KK, sep = ""))
        }
    }

    t00 <- Sys.time()
    solution.tclust <- optimalFlow::tclust_H(as.matrix(cytometry[, 1:dim.cyto]), k = KK, alpha = trimming, restr.fact = restr.fact,
        nstart = 10, iter.max = 100, equal.weights = FALSE, sol_ini = sol_inicial, sol_ini_p = TRUE, restr = "eigen", trace = 0,
        zero.tol = 1e-16)
    t11 <- Sys.time()
    t11 - t00

    t1 <- Sys.time()
    n_clus <- length(table(solution.tclust$cluster)) - 1
    print(paste("tclust found k = ", n_clus, sep = ""))
    print(paste(difftime(t1, t0), units(difftime(t1, t0))))
    cat("\n")
    return(list(cluster = solution.tclust$cluster, n_clus = n_clus, obj = solution.tclust$obj))
}

#' costWasserMatchinEllipse
#'
#' Calculates a similarity distance based on the 2-Wassertein distance between mixtures of multivariate normal distributions.
#'
#' @param test.cytometry A clusetering represented as a list of clusters. Each cluster is a list with elements mean, cov, weight and type.
#' @param training.cytometries A list of clusterings with the same format as test.cytometry.
#' @param equal.weights If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.
#'
#' @return A vector representing the similarity distance between test.cytometry and the elements in training.cytometries.
#'
#' @examples
#' partition1 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = '1'),
#'                   list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = '2'))
#' partition2 <- list(list(list(mean = c(1, -1), cov = diag(1, 2),
#'                             weight = 0.5, type = '1'), list(mean = c(-1, 1), cov = diag(1, 2), weight = 0.5, type = '2')))
#' costWasserMatchingEllipse(partition1, partition2)
#'
#' @references E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
#'
#' @export
#'
costWasserMatchingEllipse <- function(test.cytometry, training.cytometries, equal.weights = FALSE) {
    dim.test.cytometries <- length(training.cytometries)
    cost.vector <- array(0, dim = c(1, dim.test.cytometries))
    training.cytometries_elipses <- list()
    for (i in 1:(dim.test.cytometries)) {
        training.cytometry <- training.cytometries[[i]]
        names.a <- unlist(lapply(test.cytometry, function(x) x$type))
        names.b <- unlist(lapply(training.cytometry, function(x) x$type))
        weights.a <- unlist(lapply(test.cytometry, function(x) x$weight))
        if (abs(1 - sum(weights.a)) > 10^(-5)) {
            message("Weights do not add to 1. A normalization will be applied.")
            weights.a <- weights.a/sum(weights.a)
        }
        weights.b <- unlist(lapply(training.cytometry, function(x) x$weight))
        if (abs(1 - sum(weights.b)) > 10^(-5)) {
            message("Weights do not add to 1. A normalization will be applied.")
            weights.b <- weights.b/sum(weights.b)
        }

        if (equal.weights) {
            a <- rep(1/length(names.a), length(names.a))
            b <- rep(1/length(names.b), length(names.b))
            names(a) <- names.a
            names(b) <- names.b
        } else {
            a <- weights.a
            b <- weights.b
            names(a) <- names.a
            names(b) <- names.b
        }

        nn <- length(a)
        mm <- length(b)
        cost.ab <- array(dim = c(nn, mm))

        naive.transport.cost <- 0
        for (k in 1:nn) {
            for (l in 1:mm) {
                cost.ab[k, l] <- optimalFlow::w2dist(test.cytometry[[k]], training.cytometry[[l]])
                naive.transport.cost <- naive.transport.cost + cost.ab[k, l] * a[k] * b[l]
            }
        }
        tranport.plan <- transport::transport(a, b, cost.ab)
        transport.cost <- 0
        for (k in 1:dim(tranport.plan)[1]) {
            transport.cost <- transport.cost + cost.ab[tranport.plan[k, 1], tranport.plan[k, 2]] * tranport.plan[k, 3]
        }
        cost.vector[1, i] <- transport.cost/naive.transport.cost
    }
    return(cost.vector)
}

#' wasserCostFunction
#'
#' Calculates the similarity distance between elements j and i of a list of partitions.
#'
#' @param j An entry of the list of partitions.
#' @param i An entry of the list of partitions.
#' @param  cytometries The list of partitions.
#' @param equal.weights If True, weights assigned to every cluster in a partion are
#'  uniform (1/number of clusters) when calculating the similarity distance.
#'  If False, weights assigned to clusters are the proportions of points in every
#'   cluster compared to the total amount of points in the partition.
#'
#' @return A double giving the value of the similarity distance.
#'
#' @examples
#' # # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
#' database <- buildDatabase(
#'   dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
#'     population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
#'
#' templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
#' cl.paral = 1)
#' print(wasserCostFunction(1, 2, list(templates.optimalFlow$database.elliptical[[1]],
#'  templates.optimalFlow$database.elliptical[[2]])))
#'
#' @export
#'
wasserCostFunction <- function(j, i, cytometries, equal.weights = FALSE) {
    cyto_a <- cytometries[[j]]
    cyto_b <- cytometries[[i]]
    names_a <- unlist(lapply(cyto_a, function(x) x$type))
    names_b <- unlist(lapply(cyto_b, function(x) x$type))
    weights_a <- unlist(lapply(cyto_a, function(x) x$weight))
    if (abs(1 - sum(weights_a)) > 10^(-5)) {
        message("Weights do not add to 1. A normalization will be applied.")
        weights_a <- weights_a/sum(weights_a)
    }
    weights_b <- unlist(lapply(cyto_b, function(x) x$weight))
    if (abs(1 - sum(weights_b)) > 10^(-5)) {
        message("Weights do not add to 1. A normalization will be applied.")
        weights_b <- weights_b/sum(weights_b)
    }

    if (equal.weights) {
        a <- rep(1/length(names_a), length(names_a))
        b <- rep(1/length(names_b), length(names_b))
        names(a) <- names_a
        names(b) <- names_b
    } else {
        a <- weights_a
        b <- weights_b
        names(a) <- names_a
        names(b) <- names_b
    }

    nn <- length(a)
    mm <- length(b)
    cost_ab <- array(dim = c(nn, mm))

    d_w_naive <- 0
    for (k in 1:nn) {
        for (l in 1:mm) {
            cost_ab[k, l] <- optimalFlow::w2dist(cyto_a[[k]], cyto_b[[l]])
            d_w_naive <- d_w_naive + cost_ab[k, l] * a[k] * b[l]
        }
    }
    t_plan <- transport::transport(a, b, cost_ab)
    t_cost <- 0
    for (k in 1:dim(t_plan)[1]) {
        t_cost <- t_cost + cost_ab[t_plan[k, 1], t_plan[k, 2]] * t_plan[k, 3]
    }
    transport_cost <- t_cost/d_w_naive
}

#COMPATIBILITY WITH TCLUST CRAN names of input and output parameters and their corresponding equivalences

#input parameters in TCLUST CRAN and equivalence here in brackets
#x A matrix or data (x)
#k The number of clusters (k)
#alpha The proportion of observations to be trimmed (alpha)
#nstart The number of random initializations to be performed (nstart)
#iter.max The maximum number of concentration steps to be performed (iter.max)
#restr The type of restriction to be applied on the cluster scatter matrices "eigen"  "deter" and "sigma" (restr)
#restr.fact = 1.
#equal.weights specifying whether equal cluster weights are equal (equal.weights)
#zero.tol The zero tolerance used. By default set to 1e-16 (zero.tol)


#output parameters in TCLUST CRAN and equivalence here in brackets
#centers A matrix of size p x k containing the centers columwise of each cluster. (centers)
#cov An array of size p x p x k containing the covariance matrices of each cluster (cov)
#cluster A numerical vector of size n containing the cluster assignment for each observation (clusters)
#obj The value of the objective function of the best solution (obj)
#weights A numerical vector of length k, containing the weights of each cluster (weights)


#####FUNCTIONS DEVOTED TO APPLY CONSTRAINTS TO COVARIANCE MATRICES : restr2_eigen  restr2_deter_ restr.avgcov restr.diffax

##restr2_eigen
##FUNCTION FOR APPLYING EIGEN CONSTRAINTS. These are the typical constraints
## Fritz, H., Garcia-Escudero, L. A., & Mayo-Iscar, A. (2012). tclust: An r package for a trimming approach to cluster analysis. Journal of Statistical Software, 47(12), 1-26.

#' restr2_eigenv
#'
#' Function for applying eigenvalue constraints. These are the typical constraints used in tclust.
#'
#' @param autovalues Matrix containing eigenvalues.
#' @param ni.ini Current sample size of the clusters.
#' @param factor_e The level of the constraints.
#' @param zero.tol Tolerance level.
#'
#' @return The restricted eigenvalues.
#'
#' @examples
#' autovalues <- matrix(c(2, 3, 4, 1, 2, 3), nrow = 2)
#' ni.ini <- c(2, 2, 3)
#' factor_e <- 1.1
#' zero.tol <- 1e-9
#' autovalues_const <- restr2_eigenv(autovalues, ni.ini, factor_e, zero.tol)
#' autovalues_const
#'
#' @noRd
#'
restr2_eigenv <- function(autovalues, ni.ini, factor_e, zero.tol) {
###### function parameters:
###### autovalues: matrix containin eigenvalues
###### ni.ini: current sample size of the clusters
###### factor_e: the level of the constraints
###### zero.tol:toletance level

  	###### inicializations
    c <- factor_e
  	d <- t(autovalues)
    p <- nrow(autovalues)
  	K <- ncol(autovalues)
  	n <- sum(ni.ini)
  	nis <- matrix(data = ni.ini, nrow = K, ncol = p)

  	###### d_ is the ordered set of values in which the restriction objective function change the definition
  	###### points in d_ correspond to  the frontiers for the intervals in which this objective function has the same definition
  	###### ed is a set with the middle points of these intervals

  	d_ <- sort(c(d,d/c))
  	dim <- length(d_)
  	d_1 <- d_
  	d_1[dim + 1] <- d_[dim] * 2
  	d_2 <- c(0, d_)
  	ed <- (d_1 + d_2)/2
  	dim <- dim + 1;

  	###### the only relevant eigenvalues are those belong to a clusters with sample size greater than 0.
  	###### eigenvalues corresponding to a clusters whit 0 individuals has no influence in the objective function
  	###### if all the eigenvalues are 0 during the smart initialization we assign to all the eigenvalues the value 1

  	if ((max(d[nis > 0]) <= zero.tol))
  		  return(matrix(0, nrow = p, ncol = K))	        ##  solution corresponds to 0 matrix

  	###### we check if the  eigenvalues verify the restrictions

  	if (abs(max(d[nis > 0])/min(d[nis > 0])) <= c) {
    		d[nis == 0] <- mean(d[nis > 0])
    		return(t(d))					## the solution correspond to the input becuse it verifies the constraints
    		#dfin=d
  	}

  	###### we build the sol array
  	###### sol[1],sol[2],.... this array contains the critical values of the interval functions which defines the m objective function
  	###### we use the centers of the interval to get a definition for the function in each interval
  	###### this set with the critical values (in the array sol) contains the optimum m value

  	t <- s <- r <- array(0, c(K, dim))
  	sol <- sal <- array(0, c(dim))

  	for (mp_ in 1:dim) {
    		for (i in 1:K) {
      			r[i, mp_] <- sum((d[i, ] < ed[mp_])) + sum((d[i, ] > ed[mp_] * c))
      			s[i, mp_] <- sum(d[i, ] * (d[i, ] < ed[mp_]))
      			t[i, mp_] <- sum(d[i, ] * (d[i, ] > ed[mp_] * c))
    		}

    		sol[mp_] <- sum(ni.ini/n * (s[, mp_] + t[, mp_]/c))/(sum(ni.ini/n * (r[, mp_])))

    		e <- sol[mp_] * (d < sol[mp_]) + d * (d >= sol[mp_])*(d <= c * sol[mp_]) + (c * sol[mp_]) * (d > c * sol[mp_])
    		o <- -1/2 * nis/n * (log(e) + d/e)

    		sal[mp_] <- sum(o)
  	}

  	###### m is the optimum value for the eigenvalues procedure
  	eo <- which.max(c(sal))
  	m <- sol[eo]

  	###### based on the m value we get the restricted eigenvalues

  	t(m * (d < m) + d * (d >= m) * (d <= c * m) + (c * m) * (d > c * m))	##	the return value
}


#      restr.diffax
##	function which manages constraints application

#' restr.diffax
#'
#' Ment for applying restrictions to solutions obtained by tclust_.
#'
#' @param iter Current solution obtained by tclust_.
#' @param pa Parameters for using in tclust_.
#'
#' @return The new restricted solution.
#'
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                 maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                 sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' pa$maxfact_e <- 1.1
#' print(iter$sigma)
#' iter2 <- restr.diffax(iter, pa)
#' print(iter2$sigma)
#'
#' @noRd
#'
restr.diffax <- function (iter, pa) {

  	u <- array(NA, c(pa$p, pa$p, pa$K))
  	d <- array(NA, c(pa$p, pa$K))

  	for (k in 1:pa$K) {
    		ev <- eigen(iter$sigma[, , k])
    		u[, , k] <- ev$vectors
    		d[, k] <- ev$values
  	}

  	d[d < 0] <- 0		##	all eigenvalue < 0 are assigned to 0, this issue appears for numerical errors

    d <- restr2_eigenv(autovalues = d, ni.ini = iter$csize, factor_e = pa$maxfact_e, zero.tol = pa$zero.tol)
  	##	checking for singularity in all clusters.
  	iter$code <- max(d) > pa$zero.tol

  	if (!iter$code)
    		return(iter)

  	for (k in 1:pa$K)	##	re-composing the sigmas
  		  iter$sigma[, , k] <- u[, , k] %*% diag(d[, k], nrow = pa$p) %*% t(u[, , k])

  	return(iter)
}




###### MISCELANEOUS FUNCTIONS: dmnorm ssclmat TreatSingularity

##	Multivariate normal density

#' dmnorm
#'
#' The multivariate normal density.
#'
#' @param X Multivariate points.
#' @param mu Mean.
#' @param sigma Covariance matrix.
#'
#' @return The value of the multivariate normal with mean mu and covariance sigma at the points in X.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' dmnorm(X = x, mu = c(0, 0), sigma = diag(2))
#'
#' @importFrom stats mahalanobis
#'
#' @noRd
#'
dmnorm <- function (X, mu, sigma) ((2 * pi)^(-length(mu)/2)) * (det(sigma)^(-1/2)) * exp(-0.5 * mahalanobis(X, mu, sigma))


##	get a matrix object out of the sigma

#' ssclmat
#'
#' Extract a matrix from the object containing covariance matrices. Ment for use inside tclust_ function.
#'
#' @param x An object used inside tclust_ containing covariance matrices.
#' @param k An integer value giving a location in an array.
#'
#' @return A covariance matrix.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' pa$maxfact_e <- 1.1
#' iter2 <- restr.diffax(iter, pa)
#' ssclmat(iter2$sigma, k = 1)
#'
#' @noRd
#'
ssclmat <- function (x, k) as.matrix(x[, , k])

##	to manage singular situations

#' TreatSingularity
#'
#' Shows a warning message when a singularity in the curent solution of tclust_ is found.
#'
#' @param iter Current solution obtained by tclust_.
#' @param pa Parameters for using in tclust_.
#'
#' @return A warning message.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x , K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' TreatSingularity(iter, pa)
#'
#' @noRd
#'
TreatSingularity <- function (iter, pa) {
  	warning("points in the data set are concentrated in k points after trimming ")
  	return(iter)
}


######## FUNCTIONS FOR RANDOM STARTS:  getini    InitClusters
##	calculates the initial cluster sizes

#' getini
#'
#' Calculates the initial cluster sizes for initializing tclust_.
#'
#' @param K Number of groups.
#' @param no.trim Indicates if trimming is allowed.
#'
#' @return A vector of length K indicating indices of the points to be considered for initialization.
#' @examples
#' v <- getini(K = 3, no.trim = 100)
#' v
#'
#' @noRd
#'
getini <- function (K, no.trim) {
  	if (K == 1)
  		  return(no.trim)

  	pi.ini <- runif(K)
  	ni.ini <- sample(x = K, size = no.trim, replace = TRUE, prob = pi.ini/sum(pi.ini))
  	return(tabulate(ni.ini, nbins = K))
}

##	calculates the initial cluster assignment and initial values for the parameters

#' InitClusters
#'
#' Calculates the initial cluster assignment and initial values for the parameters
#'
#' @param X Points in a multidimiensional space.
#' @param iter Current solution obtained by tclust_.
#' @param Parameters for using in tclust_.
#'
#' @return An initial solution, based on a random subsample, in a form of an object used internally in tclust_.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100)+ 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16,  trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' iter <- InitClusters(X = x, iter = output3$iter, pa = output3$pa)
#' iter$cw
#' iter$center
#' iter$sigma
#'
#' @importFrom stats cov
#'
#' @noRd
#'
InitClusters <- function (X, iter, pa) {
  	dMaxVar <- 0
  	for (k in 1:pa$K) {
    		idx <- sample(1:pa$n, pa$p + 1)
    		X.ini <- X[drop = FALSE, idx, ]#sample (1:pa$n, pa$p+1),]	##	selecting observations randomly for the current init - cluster

    		iter$center[k, ] <- colMeans(X.ini)			##	calculating the center

    		cc <- (pa$p/(pa$p + 1)) * cov(X.ini)			##	calculating sigma (cc = current cov)
    		iter$sigma[, , k] <- cc
  	}
  	if (pa$equal.weights) {						##	if we're considering equal weights, cw is set here AND NEVER CHANGED
  	    iter$csize <- rep(pa$no.trim/pa$K, pa$K)
  	}	else {
  	      iter$csize = getini(pa$K, pa$no.trim)
  	}
  	iter$cw <- iter$csize/pa$no.trim				## if we're considering different weights, calculate them, and they're gonna be recalculated every time in findClustAss
  	return(iter)
}


######## FUNCTION FOR estimating model parameters: estimClustPar
##PAPERS
##HARD ASSIGNMEMT Fritz, H., Garcia-Escudero, L. A., & Mayo-Iscar, A. (2012). tclust: An r package for a trimming approach to cluster analysis. Journal of Statistical Software, 47(12), 1-26.
##MIXTURE MODEL Garc?a-Escudero, Luis Angel, Alfonso Gordaliza, and Agust?n Mayo-Iscar. "A constrained robust proposal for mixture modeling avoiding spurious solutions." Advances in Data Analysis and Classification 8.1 (2014): 27-43.
##FUZZY CLUSTERING Fritz, H., Garc?A-Escudero, L. A., & Mayo-Iscar, A. (2013). Robust constrained fuzzy clustering. Information Sciences, 245, 38-52.

#' estimClustPar
#'
#' Obtain the best values for the model parameters, given data, input parameters and an assigment.
#'
#' @param X Points in a multidimiensional space.
#' @param iter Current solution obtained by tclust_.
#' @param pa Parameters for using in tclust_.
#'
#' @return An object used internally in tclust_ with the best model parameters.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' output4 <- estimClustPar(X = x, iter, pa)
#' output4$center
#' output4$sigma
#' output4$cw
#'
#' @noRd
#'
estimClustPar <- function (X, iter, pa) {
    ###  this line must dissapear because  iter$csize  is updated in findClustAssig          iter$csize=apply(iter$z_ij,2,'sum')
		for (k in 1:pa$K) {
  			if (iter$csize[k] > pa$zero.tol) { ##	if cluster's size is > 0
    				iter$center[k, ] = (t(iter$z_ij[, k]) %*% X)/iter$csize[k]
    				X.c <- (X - matrix(iter$center[k, ], ncol = pa$p, nrow = pa$n, byrow = TRUE))
    				iter$sigma[, , k] <- (t(X.c * iter$z_ij[, k]) %*% X.c)/iter$csize[k]
  			}	else							##	this cluster's size has decreased to 0
  			    iter$sigma[, , k] <- 0
    }
	return(iter)
}


######## FUNCTIONS FOR obtaining the assigment and trimming: findClustAssig (mixture models and hard assigment)  findClustAssig_  (fuzzy assigment)

######## FUNCTION FOR obtaining the assigment and trimming in the non FUZZY CASE (mixture and hard assignments)
##PAPERS
##HARD ASSIGNMEMT Fritz, H., Garcia-Escudero, L. A., & Mayo-Iscar, A. (2012). tclust: An r package for a trimming approach to cluster analysis. Journal of Statistical Software, 47(12), 1-26.
##MIXTURE MODEL Garc?a-Escudero, Luis Angel, Alfonso Gordaliza, and Agust?n Mayo-Iscar. "A constrained robust proposal for mixture modeling avoiding spurious solutions." Advances in Data Analysis and Classification 8.1 (2014): 27-43.

#' findClustAssig
#'
#' Function for obtaining the assigment and trimming (mixture models and hard assigment).
#'
#' @param X Points in a multidimiensional space.
#' @param iter Current solution obtained by tclust_.
#' @param pa Parameters for using in tclust_.
#'
#' @return An object used internally in tclust_ with the assignment of the input points.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' output5 <- findClustAssig(X=x, iter, pa)
#'
#' @noRd
#'
findClustAssig <- function (X, iter, pa) {
    ll = matrix(NA, pa$n, pa$K)
    for (k in 1:pa$K)	ll[, k] <- iter$cw[k]*dmnorm(X, iter$center[k, ], ssclmat(iter$sigma, k)) ## dmvnorm could be used here...

  	old.assig <- iter$assig
  	iter$assig <- apply(ll, 1, which.max)						##	searching the cluster which fits best for each observation

    pre.z_h <- apply(ll, 1, 'max')

    pre.z_m <- apply(ll, 1, 'sum')

    pre.z_ <- matrix(pre.z_m, nrow = pa$n, ncol = pa$K, byrow = FALSE)

    ##### To obtain the trimming:  tc.set is the non trimming indicator

    tc.set <- (rank(pre.z_h, ties.method = "random") > floor(pa$n * (pa$alpha)))

    ##### To obtain the iter$z_ij matrix contining the assigment and trimming
    iter$assig <- apply(ll, 1, which.max)*tc.set    #hard assigment including trimming

    iter$z_ij <- 0 * iter$z_ij
    iter$z_ij[cbind((1:pa$n), iter$assig + (iter$assig == 0))] <- 1
    iter$z_ij[tc.set == FALSE, ] <- 0
    iter$code <- 2 * all(old.assig == iter$assig)		##	setting the code - parameter, signaling whether the assignment is the same than the previous --- is the only stopping rule implemented

    ##### To obtain the size of the clusters and the estimated weight of each population
  	iter$csize <- tabulate(iter$assig, pa$K)

    if (!pa$equal.weights) iter$cw <- iter$csize/sum(iter$csize)					##	and cluster weights

    return(iter)
}




######## FUNCTION FOR obtaining the objetive functions value for mixture (obj_m) hard (obj_h) and fuzzy (obj_f)
##PAPERS
##HARD ASSIGNMEMT Fritz, H., Garcia-Escudero, L. A., & Mayo-Iscar, A. (2012). tclust: An r package for a trimming approach to cluster analysis. Journal of Statistical Software, 47(12), 1-26.
##MIXTURE MODEL Garc?a-Escudero, Luis Angel, Alfonso Gordaliza, and Agust?n Mayo-Iscar. "A constrained robust proposal for mixture modeling avoiding spurious solutions." Advances in Data Analysis and Classification 8.1 (2014): 27-43.
##FUZZY CLUSTERING Fritz, H., Garc?A-Escudero, L. A., & Mayo-Iscar, A. (2013). Robust constrained fuzzy clustering. Information Sciences, 245, 38-52.

#' calcobj
#'
#' Calculates tclust's objective function value.
#'
#' @param X Points in a multidimiensional space.
#' @param iter Current solution obtained by tclust_.
#' @param pa Parameters for using in tclust_.
#'
#' @return An object used internally in tclust_ with the the value for tclust's objective function.
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2),
#'         matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' iter <- output3$iter
#' pa <- output3$pa
#' iter_ <- calcobj(X = x, iter = iter, pa = pa)
#' iter_$obj
#'
#' @noRd
#'
calcobj <- function (X, iter, pa) {
    ww_m <- matrix(0, nrow = pa$n, ncol = 1)
    ww_h <- matrix(0, nrow = pa$n, ncol = 1)

  	for (k in 1:pa$K) {
         w_m <- iter$cw[k] * dmnorm(X, iter$center[k, ], ssclmat(iter$sigma, k))
         ww_m <- w_m * (w_m >= 0) + ww_m     ##	calculates each individual contribution for the obj funct mixture
         w_h <- w_m * (iter$assig == k)
         ww_h <- w_h * (w_h >= 0) + ww_h     ##	calculates each individual contribution for the obj funct hard
    }
    ww_m <- ww_m * (ww_m >= 0)
    ww_h <- ww_h * (ww_h >= 0)

    iter$obj <- sum(log(ww_h[iter$assig > 0]))
  	return(iter)
}
#' tclust_
#'
#' Function that performs robust non spherical clustering, tclust, where initial solutions are allowed.
#'
#' @param X Points in a multidimiensional space.
#' @param K Number of clusters.
#' @param alpha Level of trimming.
#' @param niter Maximum number of iterations.
#' @param Ksteps Maximum number of K-steps.
#' @param equal.weights If all clusters should have equal weight in the objective function.
#' @param restr.cov.value Type of restriction on covariance matrices. Only "eigen" will be allowed.
#' @param maxfact_e Level determinant constraints.
#' @param zero.tol The zero tolerance used. By default set to 1e-16.
#' @param trace Defines the tracing level, which is set to 0 by default. Tracing level 2 gives additional information on the iteratively decreasing objective function's value.
#' @param sol_ini_p Initial solution for parameters provided by the user TRUE/FALSE, if TRUE is stored in sol_ini.
#' @param sol_ini Initial solution for parameters provided by the user.
#'
#' @return A list with values:
#' \describe{
#'  \item{iter}{Contains the solutions of the clustering provided by tclust_.}
#'  \item{pa}{Parametaers used in tslucs_.}
#'  \item{X}{The input variables.}
#' }
#'
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2), matrix(rnorm(100) + 4, ncol = 2))
#' output3 <- tclust_(X = x, K = 3, alpha = 0.05, niter = 20, Ksteps = 10,
#'                  equal.weights = FALSE, restr.cov.value = "eigen",
#'                  maxfact_e = 5, zero.tol = 1e-16, trace = 0,
#'                  sol_ini_p = FALSE, sol_ini = NA)
#' output3$iter$assig
#' plot(x, col = output3$iter$assig + 1)
#'
#' @noRd
#'
tclust_ <- function(X, K, alpha = 0.05, niter = 20, Ksteps = 10, equal.weights = FALSE,
                    restr.cov.value = "eigen", maxfact_e = 5, zero.tol = 1e-16, trace = 0,
                    sol_ini_p = FALSE, sol_ini = NA) {

	  if (!is.numeric(X))
		    stop("parameter x: numeric matrix/vector expected")
	  if (!is.matrix(X))
		    X <- matrix(X, ncol = 1)
  	n <- nrow(X)
  	p <- ncol(X)
  	no.trim <- floor(n * (1 - alpha))
    f.restr <- restr.diffax

  	# preparing lot's of lists: pa ( input parameters for the procedure)  iter (current value of the parameters of the model)
  	pa <- list(		            ##	 input parameters for the procedure
    		n = n,			    ##	number of observations
    		p = p,	                    ##	number of dimensions
    		alpha = alpha,		    ##	level of trimming
    		trimm = n - no.trim,	    ##	number of observations which are considered as to be outliers
    		no.trim = no.trim,	    ##	number of observations which are considered as to be not outliers
    		K = K,			    ##	number of clusters
    		equal.weights = equal.weights,		##	equal population proportions  for all clusters
    		zero.tol = zero.tol,			##	zero tolerance	(to be implemented)
    		trace = trace,				##	level of information provided (to be implemented)
        maxfact_e = maxfact_e,                    ##	level determinant constraints
        sol_ini_p = sol_ini_p,                  ##	initial solution for parameters provided by the user TRUE/FALSE   if TRUE is stored in sol_ini
        sol_ini = sol_ini,                      ##	initial solution for parameters provided by the user
        Ksteps = Ksteps,                          ##	number of iterations
        niter = niter                             ##	number of random starts
  	)

  	iter <- list(					##	current value of the parameters
    		obj = NA,				##	current objective value
    		assig = array(0, n),			##	cluster group of assignment
    		csize = array(NA, K),			##	cluster number of observations assigned to the cluster
    		cw = rep(NA, K),			##	cluster estimated proportions
    		sigma = array(NA, c (p, p, K)),	##	cluster's covariance matrix
    		center = array(NA, c(K, p)),		##	cluster's centers
    		code = NA,				##	this is a return code supplied by functions like findClustAssig
    		z_ij = matrix(0, nrow = n, ncol = K),	##	cluster assignment given by 0/1 columns
    		lambda = rep(NA, K)	##	diagonal values for tkmeans
  	)
	  best.iter <- list(obj = -Inf)			##	for containing the best values for the paramteters after several random starts

    if (pa$sol_ini_p == TRUE) pa$niter <- 1

	  for (j in 1:pa$niter) {
      if (pa$sol_ini_p == TRUE) {
        if (is.numeric(sol_ini$assig)) {
            iter$assig <- sol_ini$assig
            iter$z_ij <- 0 * iter$z_ij
            for (h in 1:pa$K) iter$z_ij[iter$assig == h, h] <- 1
  	        iter$csize <- tabulate(iter$assig, pa$K)
            iter$cw <- iter$csize/sum(iter$csize)
   	        iter <- estimClustPar(X, iter, pa)
   	        iter$center[is.na(iter$center)] <- 0
        } else  {
            iter$cw <- pa$sol_ini$cw;  iter$center <- pa$sol_ini$center; iter$sigma <- pa$sol_ini$sigma
	          iter$csize <- iter$cw*pa$n * (1 - pa$alpha)
	      }
      } else  iter <- InitClusters(X, iter, pa)   #initial solution provided by the user (pa$sol_ini_p==TRUE) or at random (FALSE)

      iter$center[is.na(iter$center)] <- 0
		  for (i in 0:pa$Ksteps) {
          iter <- f.restr(iter = iter, pa = pa)	##	restricting the clusters' scatter structure
    			if (iter$code == 0) {								##	all eigenvalues are zero
    				if (i > 0)
    					return(TreatSingularity(calcobj(X, iter, pa), pa))
    				else
    					iter$sigma[, , ] <- diag(pa$p)
          }
    		  iter <- findClustAssig(X, iter, pa)		## estimates the cluster's assigment and TRIMMING (mixture models and HARD )
    			if ((iter$code == 2) ||									##		if findClustAssig returned 1, meaning that the cluster Assignment has not changed
    		         (i == pa$Ksteps))									##		or we're in the last concentration step:
    				break											##		break the for - loop - we finished this iteration! dont re-estimate cluster parameters this time
    			iter <- estimClustPar(X, iter, pa)		## estimates the cluster's parameters
		  }
		  iter <- calcobj(X, iter, pa)							##	calculates the objetive function value
		if (iter$obj > best.iter$obj)
			best.iter <- iter
    best.iter_ <- list(iter = best.iter, pa = pa, X = X)
	}
	return(best.iter_)
}
#' tclust_H
#'
#' A wrapper for the internal fucntion tclust_. Performs robust non spherical clustering, tclust, where initial solutions are allowed.
#'
#' @param x A matrix or data.frame of dimension n x p, containing the observations (row-wise).
#' @param k The number of clusters initially searched for.
#' @param alpha The proportion of observations to be trimmed.
#' @param nstart The number of random initializations to be performed. Only when sol_ini_p = FALSE.
#' @param iter.max The maximum number of concentration steps to be performed. The concentration steps are stopped, whenever two consecutive steps lead to the same data partition.
#' @param restr The type of restriction to be applied on the cluster scatter matrices. Valid values are "eigen" (default).
#' @param restr.fact The constant restr.fact >= 1 constrains the allowed differences among group scatters. Larger values imply larger differences of group scatters, a value of 1 specifies the strongest restriction.
#' @param sol_ini_p Initial solution for parameters provided by the user TRUE/FALSE, if TRUE is stored in sol_ini.
#' @param sol_ini Initial solution for parameters provided by the user.
#' @param equal.weights A logical value, specifying whether equal cluster weights (TRUE) or not (FALSE) shall be considered in the concentration and assignment steps.
#' @param trace Defines the tracing level, which is set to 0 by default. Tracing level 2 gives additional information on the iteratively decreasing objective function's value.
#' @param zero.tol The zero tolerance used. By default set to 1e-16.
#'
#' @return A list with values:
#' \describe{
#'  \item{centers}{A matrix of size p x k containing the centers (column-wise) of each cluster.}
#'  \item{cov}{An array of size p x p x k containing the covariance matrices of each cluster.}
#'  \item{cluster}{A numerical vector of size n containing the cluster assignment for each observation. Cluster names are integer numbers from 1 to k, 0 indicates trimmed observations.}
#'  \item{par}{A list, containing the parameters the algorithm has been called with (x, if not suppressed by store.x = FALSE, k, alpha, restr.fact, nstart, KStep, and equal.weights).}
#'  \item{weights}{A numerical vector of length k, containing the weights of each cluster.}
#'  \item{obj}{he value of the objective function of the best (returned) solution.}
#' }
#' @details
#' This iterative algorithm initializes k clusters randomly and performs "concentration steps" in order to improve the current cluster assignment. The number of maximum concentration steps to be performed is given by iter.max. For approximately obtaining the global optimum, the system is initialized nstart times and concentration steps are performed until convergence or iter.max is reached. When processing more complex data sets higher values of nstart and iter.max have to be specified (obviously implying extra computation time). However, if more then half of the iterations would not converge, a warning message is issued, indicating that nstart has to be increased.
#'
#' The parameter restr defines the cluster's shape restrictions, which are applied on all clusters during each iteration. Options "eigen"/"deter" restrict the ratio between the maximum and minimum eigenvalue/determinant of all cluster's covariance structures to parameter restr.fact. Setting restr.fact to 1, yields the strongest restriction, forcing all eigenvalues/determinants to be equal and so the method looks for similarly scattered (respectively spherical) clusters. Option "sigma" is a simpler restriction, which averages the covariance structures during each iteration (weighted by cluster sizes) in order to get similar (equal) cluster scatters.
#'
#' @examples
#' x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2),
#'         matrix(rnorm(100) + 4, ncol = 2))
#' ## robust cluster obtention from a sample x asking for 3 clusters,
#' ## trimming level 0.05 and constrain level 12
#' k <- 3; alpha <- 0.05; restr.fact <- 12
#' output <- tclust_H(x = x, k = k, alpha = alpha, nstart = 50, iter.max = 20,
#'                  restr = "eigen", restr.fact = restr.fact, sol_ini_p = FALSE, sol_ini = NA,
#'                  equal.weights = FALSE, trace = 0, zero.tol = 1e-16)
#' ## cluster assigment
#' output$cluster
#' plot(x, col = output$cluster)
#'
#' @references Fritz, H., Garcia-Escudero, L. A., & Mayo-Iscar, A. (2012). tclust: An r package for a trimming approach to cluster analysis. Journal of Statistical Software, 47(12), 1-26.
#'
#' @export
#'
tclust_H <- function (x, k = 3, alpha = 0.05, nstart = 50, iter.max = 20, restr = "eigen", restr.fact = 12,
                    sol_ini_p = FALSE, sol_ini = NA, equal.weights = FALSE, trace = 0, zero.tol = 1e-16) {
  if (sol_ini_p == TRUE) {
    if (is.numeric(sol_ini$cluster)) {
      sol_ini$assig = sol_ini$cluster
      } else {
      sol_ini$cw <- sol_ini$weights; sol_ini$center <- t(sol_ini$centers); sol_ini$sigma <- sol_ini$cov
    }
  }
  tc_ <- tclust_(X = x, K = k, alpha = alpha, niter = nstart, Ksteps = iter.max, equal.weights = equal.weights, restr.cov.value = restr,
                 maxfact_e = restr.fact,  zero.tol = 1e-16,  trace = 0,
                 sol_ini_p = sol_ini_p, sol_ini = sol_ini )
  tc <- list()
  tc$centers <- t(tc_$iter$center)
  tc$cov <- tc_$iter$sigma
  tc$cluster <- tc_$iter$assig
  tc$par <- tc_$pa
  tc$weights <- tc_$iter$cw
  tc$obj <- tc_$iter$obj
  return(tc)
}

#' qdaClassification
#'
#' Gives quadratic discriminant scores to the points in data for a multivariate normal.
#'
#' @param normal A list with arguments mean, covaruance and weight.
#' @param data Data frame or matrix on which to perform qda.
#'
#' @return A score for each point.
#'
#' @examples
#' data.qda = cbind(rnorm(50), rnorm(50))
#' exp(qdaClassification(list(mean = c(0,0), cov = diag(1,2), weight = 1), data.qda))
#'
#' @export
#'
qdaClassification <- function(normal, data) {
    data <- as.matrix(data)
    media <- matrix(normal$mean, nrow = 1, ncol = length(normal$mean))
    inv <- tryCatch(solve(normal$cov), error = function(x) NA)
    if (is.na(inv[1])) {
        return(inv)
    } else {
        weight <- normal$weight
        qda_score <- -((1/2) * rowSums((data %*% inv) * data)) + data %*% inv %*% t(media) + matrix(-(1/2) * media %*% inv %*% 
            t(media) - (1/2) * log(1/det(inv)) + log(weight), nrow = dim(data)[1], ncol = 1)
        
    }
}

#' estimationCellBarycenter
#'
#' Estimates a Wasserstein barycenter for a cluster type using a collection of partitions.
#'
#' @param cell Name of the cluster of interest.
#' @param cytometries List of clusterings.
#'
#' @return A list representing the (1-)barycenter:
#' \describe{
#'  \item{mean}{Mean of the barycenter.}
#'  \item{cov}{Covariance of the barycenter.}
#'  \item{weight}{Weight associated to the barycenter.}
#'  \item{type}{Type of the cluster.}
#' }
#'
#' @examples
#' partition1 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = '1'),
#'                   list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = '2'))
#' partition2 <- list(list(mean = c(1, -1), cov = diag(1, 2), weight = 0.5, type = '1'),
#'                   list(mean = c(-1, 1), cov = diag(1, 2), weight = 0.5, type = '2'))
#' cytometries <- list(partition1, partition2)
#' estimationCellBarycenter('1',cytometries)
#'
#' @export
#'
estimationCellBarycenter = function(cell, cytometries) {
    inexes.cell <- lapply(cytometries, function(x) which(unlist(lapply(x, function(y) y$type)) == cell))
    cell.representatives <- list()
    j <- 0
    for (i in 1:length(inexes.cell)) {
        if (length(inexes.cell[[i]]) == 0) {
            next
        } else {
            j <- j + 1
            cell.representatives[[j]] <- cytometries[[i]][[inexes.cell[[i]]]]
        }
    }
    if (length(cell.representatives) > 0) {
        weight.cell <- unlist(lapply(cell.representatives, function(x) x$weight))
        cov.cell <- lapply(cell.representatives, function(x) x$cov)
        cov.barycenter.cell <- optimalFlow:::GaussianBarycenters(cov.cell, weight.cell/sum(weight.cell))$Barycenter
        mean.cell <- lapply(cell.representatives, function(x) x$mean)
        mean.barycenter.cell <- colSums(matrix(unlist(lapply(1:length(cell.representatives), function(i) mean.cell[[i]] * weight.cell[i]/sum(weight.cell))),
            nrow = length(cell.representatives), byrow = TRUE))
        weight.barycenter.cell = mean(weight.cell)
        return(list(mean = mean.barycenter.cell, cov = cov.barycenter.cell, weight = weight.barycenter.cell, type = cell))
    }
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----ej00, eval = FALSE-------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("optimalFLow")

## ----ej0, echo = TRUE, message = FALSE----------------------------------------
library(optimalFlowData)
library(optimalFlow)
library(ellipse)

## ----ej1, echo = TRUE---------------------------------------------------------
database <- buildDatabase(
 dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
   population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))

## ----ej1110, eval = FALSE-----------------------------------------------------
# templates.optimalFlow <-
#   optimalFlowTemplates(
#     database = database
#     )

## ----ej111, echo = TRUE-------------------------------------------------------
templates.optimalFlow <-
  optimalFlowTemplates(
    database = database, templates.number = 5, cl.paral = 1
    )

## ----ej1112, echo = TRUE------------------------------------------------------
length(templates.optimalFlow$templates) # The number of clusters, and, hence, of templates 
length(templates.optimalFlow$templates[[1]]) # The number of elements of the first template, it contains four cell types
templates.optimalFlow$templates[[1]][[1]] # The first element of the first template

## ----ej1113, echo = TRUE------------------------------------------------------
templates.optimalFlow$clustering

## ----ej1114, echo = TRUE------------------------------------------------------
length(templates.optimalFlow$database.elliptical) # the number of elements in the database
length(templates.optimalFlow$database.elliptical[[1]]) # the number of cell types in the first element of the database
templates.optimalFlow$database.elliptical[[1]][[1]] # the parameters corresponding to the first cell type in the first cytometry of the database 

## ----ej2, echo = TRUE---------------------------------------------------------
cytoPlotDatabase(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")

## ----ej22, eval = FALSE-------------------------------------------------------
# cytoPlotDatabase3d(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))

## ----ej3, echo = TRUE---------------------------------------------------------
cytoPlot(templates.optimalFlow$templates[[3]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")

## ----ej32, eval = FALSE-------------------------------------------------------
# cytoPlot3d(templates.optimalFlow$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))

## ----ej4, echo = TRUE---------------------------------------------------------
cytoPlotDatabase(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "", colour = FALSE)

## ----ej42, eval = FALSE-------------------------------------------------------
# cytoPlotDatabase3d(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000), colour = FALSE)

## ----ej5, echo = TRUE---------------------------------------------------------
templates.optimalFlow.barycenter <- 
  optimalFlowTemplates(
    database = database, templates.number = 5, consensus.method = "k-barycenter",
    barycenters.number = 4, bar.repetitions = 10, alpha.bar = 0.05, cl.paral = 1
    )

## ----ej52, echo = TRUE--------------------------------------------------------
templates.optimalFlow.hdbscan <- 
  optimalFlowTemplates(
    database = database, templates.number = 5, consensus.method = "hierarchical",
    cl.paral = 1
    )

## ----ej6, echo = TRUE---------------------------------------------------------
cytoPlot(templates.optimalFlow.barycenter$templates[[3]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")

## ----ej62, eval = FALSE-------------------------------------------------------
# cytoPlot3d(templates.optimalFlow.barycenter$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))

## ----ej7, echo=TRUE-----------------------------------------------------------
cytoPlot(templates.optimalFlow.hdbscan$templates[[3]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")

## ----ej72, eval = FALSE-------------------------------------------------------
# cytoPlot3d(templates.optimalFlow.hdbscan$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))

## ----ej77, echo = TRUE--------------------------------------------------------
templates.optimalFlow.unsup <-
  optimalFlowTemplates(
    database = database, hclust.method = "hdbscan", cl.paral = 1, consensus.method = "hierarchical"
    )
print(templates.optimalFlow.unsup$clustering)
print(templates.optimalFlow$clustering)
cytoPlot(templates.optimalFlow.unsup$templates[[5]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")

## ----ej80, echo = TRUE--------------------------------------------------------
test.cytometry <- Cytometry1[which(match(Cytometry1$`Population ID (name)`, c("Monocytes", "CD4+CD8-", "Mature SIg Kappa", "TCRgd-"), nomatch = 0) > 0), ]

## ----ej8, echo = TRUE---------------------------------------------------------
classification.optimalFlow <- 
  optimalFlowClassification(
    test.cytometry[, 1:10], database, templates.optimalFlow, 
    consensus.method = "pooling", cl.paral = 1
    )

## ----ej82, echo = TRUE--------------------------------------------------------
head(classification.optimalFlow$cluster)
table(classification.optimalFlow$cluster)

## ----ej83, echo = TRUE--------------------------------------------------------
length(classification.optimalFlow$clusterings)
table(classification.optimalFlow$clusterings[[1]]$cluster)

## ----ej84, echo = TRUE--------------------------------------------------------
classification.optimalFlow$assigned.template.index
templates.optimalFlow$clustering

## ----ej85, echo = TRUE--------------------------------------------------------
scoreF1.optimalFlow <- optimalFlow::f1Score(classification.optimalFlow$cluster,
                                            test.cytometry, noise.types)
print(scoreF1.optimalFlow)

## ----ej9, echo = TRUE---------------------------------------------------------
classification.optimalFlow.barycenter <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow.barycenter, consensus.method = "k-barycenter", cl.paral = 1
    )

## ----ej92, echo = TRUE--------------------------------------------------------
table(classification.optimalFlow.barycenter$cluster)
classification.optimalFlow.barycenter$cluster.vote

## ----ej922, echo = TRUE-------------------------------------------------------
classification.optimalFlow.barycenter$assigned.template.index
templates.optimalFlow.barycenter$clustering

## ----ej93, echo = TRUE--------------------------------------------------------
scoreF1.optimalFlow.barycenter <- 
  f1ScoreVoting(
    classification.optimalFlow.barycenter$cluster.vote, classification.optimalFlow.barycenter$cluster,
    test.cytometry,
    1.01, noise.types
    )
print(scoreF1.optimalFlow.barycenter$F1_score)

## ----ej10, echo = TRUE--------------------------------------------------------
classification.optimalFlow.hdbscan <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow.hdbscan, consensus.method = "hierarchical", cl.paral = 1
    )
table(classification.optimalFlow.hdbscan$cluster)
classification.optimalFlow.hdbscan$cluster.vote
classification.optimalFlow.hdbscan$assigned.template.index
templates.optimalFlow.hdbscan$clustering
scoreF1.optimalFlow.hdbscan <-
  f1ScoreVoting(
    classification.optimalFlow.hdbscan$cluster.vote, classification.optimalFlow.hdbscan$cluster,
   test.cytometry,
    1.01, noise.types
  )
print(scoreF1.optimalFlow.hdbscan$F1_score)

## ----ej11, echo = TRUE--------------------------------------------------------
classification.optimalFlow.2 <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow, consensus.method = "pooling", classif.method = "matching",
    cost.function = "ellipses", cl.paral = 1
    )
table(classification.optimalFlow.2$cluster)
table(classification.optimalFlow.2$clusterings[[1]]$cluster)
classification.optimalFlow.2$cluster.vote
classification.optimalFlow.2$assigned.template.index
templates.optimalFlow$clustering
scoreF1.optimalFlow.2 <-
  f1ScoreVoting(
    classification.optimalFlow.2$cluster.vote, classification.optimalFlow.2$cluster,
    test.cytometry,
    1.01, noise.types
    )
print(scoreF1.optimalFlow.2$F1_score)

## ----ej12, echo = TRUE--------------------------------------------------------
classification.optimalFlow.3 <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow, consensus.method = "pooling",
    classif.method = "random forest", cl.paral = 1
    )
table(classification.optimalFlow.3$cluster)
classification.optimalFlow.3$assigned.template.index # the cytometry used for learning belongs to the cluster labelled as 1 and is the first of the cytometries in that cluster, hence it is the first cytometry in the database.
templates.optimalFlow$clustering
scoreF1.optimalFlow.3 <-
  optimalFlow::f1Score(classification.optimalFlow.3$cluster,
    test.cytometry,
    noise.types
    )
print(scoreF1.optimalFlow.3)

#' f1ScoreVoting
#'
#' Calculates the F1 score fore each group in a partition, when provided with a fuzzy classification.
#'
#' @param voting A list where each entry is a vote on the respective label.
#' @param clustering Labels of the partition.
#' @param cytometry Data of the clustering, where the last variable contains the original labels.
#' @param nivel_sup level of tolerance for assigning a hard clustering. Should be greater or equal than 1. Class A is assigned if class A > nivel_sup * Class B.
#' @param noise.cells An array of labels to be considered as noise.
#'
#' @return A matrix where the first row is the F1 score, the second row is the Precision and the third row is the Recall.
#'
#' @examples
#' # # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
#' database <- buildDatabase(
#'   dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
#'     population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
#'
#' templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
#' cl.paral = 1)
#'
#' classification.optimalFlow <- optimalFlowClassification(as.data.frame(Cytometry1)[
#'  which(match(Cytometry1$`Population ID (name)`, c('Monocytes', 'CD4+CD8-',
#'                                                  'Mature SIg Kappa', 'TCRgd-'), nomatch = 0) > 0), 1:10], database, templates.optimalFlow,
#'  classif.method = 'matching', cost.function = 'ellipses', cl.paral = 1)
#'
#' f1ScoreVoting(classification.optimalFlow$cluster.vote, classification.optimalFlow$cluster,
#'               as.data.frame(Cytometry1)[which(match(Cytometry1$`Population ID (name)`,
#'                                                     c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'), nomatch = 0) > 0), ], 1.01, noise.types)
#' @import optimalFlowData
#' @import rlang
#' @import dplyr
#'
#' @export
#'
f1ScoreVoting <- function(voting, clustering, cytometry, nivel_sup, noise.cells) {
    correspondance <- array(dim = length(voting))
    clusters.to.eliminate <- sort(unique(clustering))[-which(sort(unique(clustering)) %in% names(voting))]
    if (length(clusters.to.eliminate) > 0) {
        indexes.to.eliminate <- which(clustering %in% clusters.to.eliminate)
        clustering <- clustering[-indexes.to.eliminate]
        cytometry <- cytometry[-indexes.to.eliminate, ]
    }
    dim.cyto <- dim(cytometry)[2]
    for (i in 1:length(voting)) {
        vote_new <- voting[[i]]
        noise <- which(vote_new$cell %in% noise.cells)
        if (length(noise) > 0) {
            if (length(noise) == length(vote_new$cell)) {
                correspondance[i] <- "noise"
            } else {
                ruido_s <- sum(vote_new$compound.proportion[noise])
                vote_new <- vote_new[-noise, ]
                if (ruido_s > nivel_sup * vote_new[1, 2]) {
                  correspondance[i] <- "noise"
                } else {
                  if (vote_new[1, 2] < nivel_sup * ruido_s) {
                    correspondance[i] <- paste("undecided", vote_new$cell[1], "noise")
                  } else {
                    if (dim(vote_new)[1] > 2) {
                      if (vote_new[1, 2] < nivel_sup * vote_new[2, 2]) {
                        correspondance[i] <- paste("undecided", vote_new$cell[1], vote_new$cell[2])
                      } else {
                        correspondance[i] <- paste(vote_new$cell[1])
                      }
                    } else {
                      correspondance[i] <- paste(vote_new$cell[1])
                    }
                  }
                }
            }

        } else {
            if (dim(vote_new)[1] > 2) {
                if (vote_new[1, 2] < nivel_sup * vote_new[2, 2]) {
                  correspondance[i] <- paste("undecided", vote_new$cell[1], vote_new$cell[2])
                } else {
                  correspondance[i] <- paste(vote_new$cell[1])
                }
            } else {
                correspondance[i] <- paste(vote_new$cell[1])
            }
        }
    }
    t <- 0
    aciertos <- array(dim = length(voting))
    for (j in names(voting)) {
        t <- t + 1
        if (correspondance[t] == "noise") {
            aciertos[t] <- sum(dplyr::pull(cytometry, dim.cyto)[clustering == j] %in% noise.cells)/length(dplyr::pull(cytometry,
                dim.cyto)[clustering == j])
        } else {
            aciertos[t] <- sum(dplyr::pull(cytometry, dim.cyto)[clustering == j] %in% correspondance[t])/length(dplyr::pull(cytometry,
                dim.cyto)[clustering == j])
        }
    }

    p <- array(dim = c(1, length(unique(correspondance))))
    r <- p
    t <- 0
    for (i in unique(correspondance)) {
        t <- t + 1
        if (i == "noise") {
            p[t] <- sum((aciertos * table(clustering))[correspondance %in% "noise"])/sum(table(clustering)[correspondance %in%
                "noise"])
            r[t] <- sum((aciertos * table(clustering))[correspondance %in% "noise"])/sum(dplyr::pull(cytometry, dim.cyto) %in%
                noise.cells)
        } else {
            p[t] <- sum((aciertos * table(clustering))[correspondance %in% i])/sum(table(clustering)[correspondance %in% i])
            r[t] <- sum((aciertos * table(clustering))[correspondance %in% i])/sum(dplyr::pull(cytometry, dim.cyto) %in% i)
        }
    }
    F1_score <- 2 * p * r/(p + r)
    F1_score_wass <- rbind(F1_score, p, r)
    cels_original <- names(table(dplyr::pull(cytometry, dim.cyto)))[which(table(dplyr::pull(cytometry, dim.cyto)) > 0)]
    cel_no_encontradas <- cels_original[!(cels_original %in% correspondance) & !(cels_original %in% noise.cells)]
    F1_score_wass <- cbind(F1_score_wass, matrix(0, ncol = length(cel_no_encontradas), nrow = 3))
    colnames(F1_score_wass) <- c(unique(correspondance), cel_no_encontradas)
    rownames(F1_score_wass) <- c("F1-score", "Precision", "Recall")
    return(list(F1_score = F1_score_wass, correspondance = correspondance))
}

#' estimCovCellGeneral
#'
#' Estimation of mean and covariance for a label in a partition.
#'
#' @param cell Labell of the clsuter of interest.
#' @param cytometry Data of the partition, without labels.
#' @param labels Labels of the partition.
#' @param type How to estimate covariance matrices of a cluster. 'standard' is for using cov(), while 'robust' is for using robustbase::covMcd.
#' @param alpha Only when type = 'robust'. Indicates the value of alpha in robustbase::covMcd.
#'
#' @return A list containing:
#' \describe{
#'  \item{mean}{Mean of the cluster.}
#'  \item{cov}{Covariance of the cluster.}
#'  \item{weight}{Weight associated to the cluster.}
#'  \item{type}{Type of the cluster.}
#' }
#'
#' @examples
#' estimCovCellGeneral('Basophils', Cytometry1[,1:10], Cytometry1[,11])
#'
#' @importFrom stats cov
#'
#' @export
#'
estimCovCellGeneral <- function(cell, cytometry, labels, type = "standard", alpha = 0.85) {
    indices <- which(labels == cell)
    ddd <- dim(cytometry)
    size <- length(indices)
    weight <- size/ddd[1]
    cytometry <- cytometry[indices, 1:(ddd[2])]
    if (size >= 2) {
        if (type == "standard") {
            return(list(mean = colMeans(cytometry), cov = cov(cytometry), weight = weight, type = cell))
        } else {
            if (type == "robust" & alpha * size > (ddd[2] + 1)) {
                rob_est <- tryCatch(robustbase::covMcd(cytometry, alpha = alpha), error = function(x) NA)
                if (is.na(rob_est)[1]) {
                  return(NA)
                } else {
                  return(list(mean = rob_est$center, cov = rob_est$cov, weight = weight, type = cell))
                }
            } else {
                return(NA)
            }
        }
    } else {
        return(NA)
    }
}

#' labelTransferEllipse
#'
#' Label transfer between a test partition and a training partitions viewed as a mixture of gaussians.
#'
#' @param i A dummy variable, should be any integral. Ment for use with lapply.
#' @param test.cytometry.ellipses A test clustering viewed as a mixture of multivariate normal distributions.
#' @param training.cytometries.barycenter A training partition viewed as a mixture of multivariate normal distributions.
#' @param equal.weights If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.
#'
#' @return A fuzzy relabeling consistent of a transportation plan.
#'
#' @examples
#' partition1 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = '1'),
#'                   list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = '2'))
#' partition2 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = 'a'),
#'                   list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = 'b'))
#' labelTransferEllipse(1, partition2, partition1)
#' @references E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
#'
#' @export
#'
labelTransferEllipse <- function(i, test.cytometry.ellipses, training.cytometries.barycenter, equal.weights = FALSE) {
    N <- length(training.cytometries.barycenter)
    M <- length(test.cytometry.ellipses)
    cost.matrix <- array(dim = c(N, M))
    for (j in 1:N) {
        for (i in 1:M) {
            cost.matrix[j, i] <- w2dist(list(mean = test.cytometry.ellipses[[i]]$mean, cov = test.cytometry.ellipses[[i]]$cov),
                list(mean = training.cytometries.barycenter[[j]]$mean, cov = training.cytometries.barycenter[[j]]$cov))
        }
    }
    names.a <- unlist(lapply(training.cytometries.barycenter, function(x) x$type))
    names.b <- unlist(lapply(test.cytometry.ellipses, function(x) x$type))

    if (equal.weights) {
        A <- rep(1/N, N)
        B <- rep(1/M, M)
        names(A) <- names.a
        names(B) <- names.b
    } else {
        A <- unlist(lapply(training.cytometries.barycenter, function(x) x$weight))
        A <- A/sum(A)
        names(A) <- names.a
        B <- unlist(lapply(test.cytometry.ellipses, function(x) x$weight))
        names(B) <- names.b
    }
    optimal.transport.form.A.to.B <- transport::transport(a = A, b = B, costm = cost.matrix)
    optimal.transport.form.A.to.B$to <- names(B)[optimal.transport.form.A.to.B$to]
    optimal.transport.form.A.to.B$from <- names(A)[optimal.transport.form.A.to.B$from]
    return(optimal.transport.form.A.to.B)
}

#' trimmedKBarycenter
#'
#' Calculates a 2-Wasserstein k-barycenter of a list of multivariate normal distributions.
#'
#' @param k Number k of elements in the k-barycenter.
#' @param alpha0 Level of trimming.
#' @param type.ini of initialization in c('rnd', 'plus-plus'). 'rnd' makes the common random
#' initilaization while 'plus-plus' initializes in a similar fashion to k-means++.
#' @param reps.list List of multivariate normals for which the trimmed k-barycenter should be performed.
#'
#' @return A list with values:
#' \describe{
#'  \item{variacion_wasser}{A double giving the Waserstein variation.}
#'  \item{baricentro}{A list of k elements, each of which is a member of the k-barycenter.
#'   Each eement is a normal distribution characterized by a mean and a covariance.}
#'  \item{cluster}{The assignment of the original entries to each member of the k-barycenter.}
#' }
#'
#' @examples
#' normals <- list(list(mean = c(1, 1), cov = diag(2, 2)), list(mean = c(1, 1),cov = diag(1, 2)),
#'  list(mean = c(3, 3), cov = diag(1, 2)))
#' trimmedKBarycenter(2, 0, 'rnd', normals)
#'
#' @export
#'
trimmedKBarycenter <- function(k, alpha0, type.ini = "rnd", reps.list) {
    if (type.ini == "rnd") {
        indices_prov <- sample(1:(length(reps.list)), k, replace = FALSE)
        indices_prov <- sort(indices_prov)
        bar.prov <- reps.list[indices_prov]
    } else {
        if (type.ini == "plus-plus") {
            length_rep_lista <- length(reps.list)
            j <- 1
            c0 <- sample(1:(length_rep_lista), 1)
            centers <- list(reps.list[[c0]])
            
            if (k > 1) {
                cent_pos <- (1:(length_rep_lista))[-c0]
                dist_w2 <- array(0, dim = c(k, length_rep_lista))
                dist_w2[1, ] <- abs(unlist(lapply(reps.list, distGaussian, centers[[1]])))
                D2 <- dist_w2[1, ][-c0]/sum(dist_w2[1, ][-c0])
                
                for (j in 2:k) {
                  c0 <- sample(cent_pos, 1, prob = D2)
                  c0_pos <- which(cent_pos == c0)
                  cent_pos <- cent_pos[-c0_pos]
                  centers[[j]] <- reps.list[[c0]]
                  dist_w2[j, ] <- unlist(lapply(reps.list, distGaussian, centers[[j]]))
                  s <- 0
                  D2 <- array(0, dim = length(cent_pos))
                  for (i in cent_pos) {
                    s <- s + 1
                    d_min <- min(abs(dist_w2[1:j, i]))
                    D2[s] <- d_min
                  }
                  D2 <- D2/sum(D2)
                }
            }
            bar.prov <- centers
        } else {
            return("Inicialization not recognised")
        }
    }
    
    c.asig <- trimmedMinDist(reps.list, bar.prov, alpha = alpha0)
    update.step <- kcenter(reps.list, k, c.asig)
    new.bar <- update.step$kcenters
    Trimmed.Variation.Prov <- update.step$t.variation
    Trimmed.Variation.Old <- Trimmed.Variation.Prov
    bar.prov <- new.bar
    
    diferencia <- 10
    iter <- 0
    while (diferencia > 1e-07) {
        iter <- iter + 1
        c.asig <- trimmedMinDist(reps.list, bar.prov, alpha = alpha0)
        update.step <- kcenter(reps.list, k, c.asig)
        new.bar <- update.step$kcenters
        Trimmed.Variation.Prov <- update.step$t.variation
        diferencia <- Trimmed.Variation.Old - Trimmed.Variation.Prov
        Trimmed.Variation.Old <- Trimmed.Variation.Prov
        bar.prov <- new.bar
    }
    Trimmed.Variation <- Trimmed.Variation.Prov
    list(variacion_wasser = Trimmed.Variation, baricentro = bar.prov, cluster = c.asig)
}

#' f1Score
#'
#' Calculates the F1 score fore each group in a partition.
#'
#' @param clustering The labels of the new classification.
#' @param cytometry Data of the clustering, where the last variable contains the original labels.
#' @param noise.cells An array of labels to be considered as noise.
#'
#' @return A matrix where the first row is the F1 score, the second row is the Precision and the third row is the Recall.
#'
#' @examples
#' f1Score(dplyr::pull(Cytometry3[c(sample(1:250,250),251:(dim(Cytometry3)[1])),],11),
#'         Cytometry3, noise.types)
#'
#' @import optimalFlowData
#' @import dplyr
#'
#' @references E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
#'
#' @export
#'
f1Score <- function(clustering, cytometry, noise.cells) {
    dim.cyto <- dim(cytometry)[2]
    prediction_qda_19_con_18 <- as.character(clustering)
    prediction_qda_19_con_18[prediction_qda_19_con_18 %in% noise.cells] <- rep("noise", sum(prediction_qda_19_con_18 %in% noise.cells))

    clustering <- prediction_qda_19_con_18
    lev_qda <- unique(clustering)
    p <- array(dim = c(1, length(lev_qda)))
    r <- p
    t <- 0
    for (i in lev_qda) {
        t <- t + 1
        if (i == "noise") {
            p[t] <- sum(clustering == i & dplyr::pull(cytometry, dim.cyto) %in% noise.cells)/sum(clustering == i)
            r[t] <- sum(clustering == i & dplyr::pull(cytometry, dim.cyto) %in% noise.cells)/sum(dplyr::pull(cytometry, dim.cyto) %in%
                noise.cells)
        } else {
            p[t] <- sum(clustering == i & dplyr::pull(cytometry, dim.cyto) %in% i)/sum(clustering == i)
            r[t] <- sum(clustering == i & dplyr::pull(cytometry, dim.cyto) %in% i)/sum(dplyr::pull(cytometry, dim.cyto) %in% i)
        }
    }
    F1_score <- 2 * p * r/(p + r)
    F1_score_tot <- rbind(F1_score, p, r)
    cels_original <- names(table(dplyr::pull(cytometry, dim.cyto)))[which(table(dplyr::pull(cytometry, dim.cyto)) > 0)]
    cel_no_encontradas <- cels_original[!(cels_original %in% prediction_qda_19_con_18) & !(cels_original %in% noise.cells)]
    F1_score_tot <- cbind(F1_score_tot, matrix(0, ncol = length(cel_no_encontradas), nrow = 3))
    colnames(F1_score_tot) <- c(lev_qda, cel_no_encontradas)
    rownames(F1_score_tot) <- c("F1-score", "Precision", "Recall")
    return(F1_score_tot)
}

---
title: "optimalFlow: optimal-transport approach to Flow Cytometry analysis"
author:
- name: Hristo Inouzhe
  affiliation: Universidad de Valladolid, Spain
  email: hristo.inouzhe@gmail.com
date: "`r format(Sys.time(), '%d %B, %Y')`"
output: 
    BiocStyle::html_document
vignette: >
    %\VignetteIndexEntry{optimalFlow: optimal-transport approach to Flow Cytometry analysis}
    %\VignetteEngine{knitr::rmarkdown}
    \usepackage[utf8]{inputenc}
references:
- id: optimalFlow
  title: 'optimalFlow: optimal-transport approach to Flow Cytometry analysis'
  author:
  - family: del Barrio
    given: Eustasio
  - family: Inouzhe
    given: Hristo
  - family: Loubes
    given: Jean-Michel
  - family: Mayo-Iscar
    given: Agustin
  - family: Matran
    given: Carlos
  URL: 'https://arxiv.org/abs/1907.08006'
  type: article-journal
  issued:
    year: 2019
    month: 7
- id: tclust
  title: A general trimming approach to robust cluster analysis.
  author:
  - family: Garcia-Escudero
    given: Luis-Angel
  - family: Gordaliza
    given: Alfonso
  - family: Matran
    given: Carlos
  - family: Mayo-Iscar
    given: Agustin
  container-title: The Annals of Statistics
  volume: 36
  page: 1324-1345
  type: article-journal
  issued:
    year: 2008
---



```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

# Introduction

*optimalFlow* is a package dedicated to applying optimal-transport techniques to supervised flow cytometry gating based on the results in @optimalFlow.

We provide novel methods for grouping (clustering) gated cytometries. By clustering a set of cytometries we are producing groups (clusters) of cytometries that have lower variability than the whole collection.  This in turn allows to improve greatly the performance of any supervised learning procedure. Once we have a partition (clustering) of a collection of cytometries, we provide several methods for obtaining an artificial cytometry (prototype, template) that represents in some optimal way the cytometries in each respective group. These prototypes can be used, among other things, for matching populations between different cytometries. Even more, a procedure able to group similar cytometries could help to detect individuals with a common particular condition, for instance some kind of disease.

*optimalFlowTemplates* is our procedure for clustering cytometries and obtaining templates. It is based on recent developments in the field of optimal transport such as a *similarity distance* between clusterings and a *barycenter* (Frechet mean) and *k-barycenters* of probability distributions.

We introduce *optimalFlowClassification*, a supervised classification tool for the case when a database of gated cytometries is available. The procedure uses the prototypes obtained by *optimalFlowTemplates* on the database. These are used to initialize *tclust*, a robust extension of k-means that allows for non-spherical shapes, for gating a new cytometry (see @tclust). By using a similarity distance between the best clustering obtained by *tclust* and the artificial cytometries provided by *optimalFlowTemplates* we can assign the new cytometry to the most similar template (and the respective group of cytometries). We provide several options of how to assign cell types to the new cytometry using the most relevant information, represented by the assigned template and the respective cluster of cytometries.

# Installation

Installation procedure:

```{r ej00, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("optimalFLow")
```

# optimalFlowTemplates

```{r ej0, echo = TRUE, message = FALSE}
library(optimalFlowData)
library(optimalFlow)
library(ellipse)
```
We start by providing a database of gated cytometries. In this case we select as a learning set 15 cytometries of healthy individuals, from the data provided in optimalFlowData. We will use Cytometry1 to test the results of our procedures. For simplicity and for the sake of a good visualisation we will select only some of the cell types, in particular a subset of 4 cell types. 
```{r ej1, echo = TRUE}
database <- buildDatabase(
 dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
   population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
```
Then we apply optimalFlowTemplates to obtain a clustering of the database and a template cytometry for each group. 
```{r ej1110, eval = FALSE}
templates.optimalFlow <-
  optimalFlowTemplates(
    database = database
    )
```
When running the default mode for optimalFlowTemplates we obtain a plot as in the figure bellow and then we are asked how many clusters we want to look for.
![Figure 0: learning clustering](hclust_learning.png)
From the plot it seems reasonable to look for 5 clusters of cytometries and we could introduce 5 and press enter, and the procedure will give us a clustering of the learning database and the respective templates. Since this is hard to show in a vignette, an equivalent way of doing this procedure is to execute the command bellow, where we ask for 5 clusters directly.
```{r ej111, echo = TRUE}
templates.optimalFlow <-
  optimalFlowTemplates(
    database = database, templates.number = 5, cl.paral = 1
    )
```
Now let us understand what does optimalFlowTemplates return. In the entry templates we have the artificial cytometries, viewed as mixtures of multivariate normal distributions, corresponding to the clustering of the cytometries in the database argument.
```{r ej1112, echo = TRUE}
length(templates.optimalFlow$templates) # The number of clusters, and, hence, of templates 
length(templates.optimalFlow$templates[[1]]) # The number of elements of the first template, it contains four cell types
templates.optimalFlow$templates[[1]][[1]] # The first element of the first template
```
In the argument clustering we have the clustering of the cytometries in the database argument.
```{r ej1113, echo = TRUE}
templates.optimalFlow$clustering
```
In the argument database.elliptical we have a list containing each cytometry in the database viewed as a mixture distribution. Each element of the list is a cytometry viewed as a mixture.
```{r ej1114, echo = TRUE}
length(templates.optimalFlow$database.elliptical) # the number of elements in the database
length(templates.optimalFlow$database.elliptical[[1]]) # the number of cell types in the first element of the database
templates.optimalFlow$database.elliptical[[1]][[1]] # the parameters corresponding to the first cell type in the first cytometry of the database 
```
In order to get some intuition about our methodology we are going to give some visual examples. Users can do it for their own data once they have applied optimalFlowTemplates.

We start with a two-dimensional representation of the cytometries of the cluster labelled as 3. As we have gated cytometries in the database we know every cell type, and, even more, we can consider every cytometry as a mixture of multivariate Gaussian distributions and this is stored in templates.optimalFlow$database.elliptical. The user just has to select the variables in which to project the cytometries through the variable dimensions.
```{r ej2, echo = TRUE}
cytoPlotDatabase(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")
```
Black ellipses correspond to the cell type CD4+CD8- in each cytometry and enclose 95% of the probability for the respective multivariate normal distributions. Red ellipses correspond to Mature Sig Kappa and so on.

A three-dimensional plot of the same case is provided as a static image and can be obtained using the following code.
```{r ej22, eval = FALSE}
cytoPlotDatabase3d(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))
```
![Figure 1: pooling database](database_pooling.png)

optimalFlowTemplates provides a template cytometry for each cluster, stored in the entry templates. We present here how to visualize in 2d the consensus cytometry, the template, corresponding to cluster 3. Recall that the cytometries belonging to cluster 3 have been plotted above. The code is straightforward, we access templates in templates.optimalFlow and select the third element of the list, since we are interested in cluster 3.
```{r ej3, echo = TRUE}
cytoPlot(templates.optimalFlow$templates[[3]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")
```
A three dimensional plot of the same case is provided as a static image and can be obtained using the following code.
```{r ej32, eval = FALSE}
cytoPlot3d(templates.optimalFlow$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))
```
![cosa](template_pooling.png)

It is clear that the prototype cytometry represents well the geometric information of the respective group of cytometries. This visualisation schemes allow users to check by themselves if the templates that they are obtaining are satisfying and if their clusters are really homogenous.

Another relevant situation in flow cytometry is when gatings of cytometries are available but without the identification of each cell type. For the cytometries of cluster 3 it is like we forgot about the colour, since now we do not have cell types identified. 
```{r ej4, echo = TRUE}
cytoPlotDatabase(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "", colour = FALSE)
```
The respective 3d static image can be obtained using the following code.
```{r ej42, eval = FALSE}
cytoPlotDatabase3d(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000), colour = FALSE)
```
![Figure 2: normal database](database_normal.png)
From a visual inspection there is enough geometrical information that allows us to differentiate the cel types. It is just a matter of how to capture it.

Indeed, using some unsupervised procedure to obtain the consensus element (the prototype cytometry) should be enough to capture the relevant cluster structure. This can be achieved using *otpimalFlowTemplates* as follows.

In the following chunk of code we are using optimalFlowTemplates on our database, we are again looking for 5 clusters with the default clustering procedure which is hierarchical complete-linkage but now we vary the consensus.method variable. We are selecting to obtain the template cytometry using k-barycenters in the Wasserstein space, where k is set to be 4.
```{r ej5, echo = TRUE}
templates.optimalFlow.barycenter <- 
  optimalFlowTemplates(
    database = database, templates.number = 5, consensus.method = "k-barycenter",
    barycenters.number = 4, bar.repetitions = 10, alpha.bar = 0.05, cl.paral = 1
    )
```
A different way of obtaining the consensus cytometry is to use density based hierarchical clustering, in this case hdbscan, setting consensus.method = "hierarchical". The advantage of this is that the number of cell types in the template cytometry is selected automatically.
```{r ej52, echo = TRUE}
templates.optimalFlow.hdbscan <- 
  optimalFlowTemplates(
    database = database, templates.number = 5, consensus.method = "hierarchical",
    cl.paral = 1
    )
```
As before, we can check how well the prototypes represent the group of cytometries. Again, we work with cluster 3. A 2d plot of the prototype cytometry obtained when using a 4-barycenter is provided, where colours represent different groups.
```{r ej6, echo = TRUE}
cytoPlot(templates.optimalFlow.barycenter$templates[[3]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")
```
A three dimensional plot of the same case is provided as a static image and can be obtained using the following code.
```{r ej62, eval = FALSE}
cytoPlot3d(templates.optimalFlow.barycenter$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))
```
![cosa](template_barycenter.png)

We do the same for the density based hierarchical clustering.
```{r ej7, echo=TRUE}
cytoPlot(templates.optimalFlow.hdbscan$templates[[3]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")
```
```{r ej72, eval = FALSE}
cytoPlot3d(templates.optimalFlow.hdbscan$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000))
```
![cosa](template_hdbscan.png)
From the visual representations of above we see that the different methods for obtaining a prototype cytometry for the cluster of cytometries labelled as 3 are returning similar results. Even more, results do seem to summarize in a reasonable way the information contained in the cytometries of cluster 3. Hence, doing some 2 or 3-dimensional visual inspection of the results is advisable for the user and it allows for an informed inspection of our procedures.


A totally unsupervised way of obtaining groups and templates is given by using density-based hierarchical clustering both when clustering the database of cytometries and when obtaining the prototype cytometry.
```{r ej77, echo = TRUE}
templates.optimalFlow.unsup <-
  optimalFlowTemplates(
    database = database, hclust.method = "hdbscan", cl.paral = 1, consensus.method = "hierarchical"
    )
print(templates.optimalFlow.unsup$clustering)
print(templates.optimalFlow$clustering)
cytoPlot(templates.optimalFlow.unsup$templates[[5]], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), xlab = "", ylab = "")
```

# optimalFlowClassification
Once we have a grouped database with prototype cytometries for each group we can apply different supervised classification procedures to classify a new ungated cytometry.

We start by selecting a test cytometry which we will treat as a cytometry that we want to classify in order to see how our supervised classification methods work.
```{r ej80, echo = TRUE}
test.cytometry <- Cytometry1[which(match(Cytometry1$`Population ID (name)`, c("Monocytes", "CD4+CD8-", "Mature SIg Kappa", "TCRgd-"), nomatch = 0) > 0), ]
```
Let us begin with, essentially, the default method for using optimalFlowClassification. It consists in doing quadratic discriminant analysis using the most similar template.
```{r ej8, echo = TRUE}
classification.optimalFlow <- 
  optimalFlowClassification(
    test.cytometry[, 1:10], database, templates.optimalFlow, 
    consensus.method = "pooling", cl.paral = 1
    )
```
In order to execute properly our supervised classifications we need to provide the new cytometry we want to classify, in our case it is test.cytometry without the label information. Then we need to provide the database in which we have applied optimalFlowTemplates, the object returned by optimalFlowTemplates and the consensus.method that we have used in optimalFlowTemplates. This is necesary in order to be able to perform correctly the classification task.

The result we obtain is a list that we will analyse here to make the user familiar with it. The first argument is a clustering of the cytometry of interest.
```{r ej82, echo = TRUE}
head(classification.optimalFlow$cluster)
table(classification.optimalFlow$cluster)
```
The argument clusterings contains the initial unsupervized or semi-supervized clusterings of the cytometry of interest. It is itself a list that can have as much entries as the number of templates in the semi-supervized case, or only one entry in the case of initial.method = "unsupervized". The relevant argument for the clusterings is cluster.
```{r ej83, echo = TRUE}
length(classification.optimalFlow$clusterings)
table(classification.optimalFlow$clusterings[[1]]$cluster)
```
As we see the initial clustering is not the same as the final result.

Finally, we have an entry that indicates which prototype is the closest to the new cytometry. This information is relevant since it is the prototype used for classifying in the default execution of optimalFlowClassification.
```{r ej84, echo = TRUE}
classification.optimalFlow$assigned.template.index
templates.optimalFlow$clustering
```
In this case test.cytometry is closest to the template corresponding to cluster 1 in templates.optimalFlow$clustering.

As we are performing supervised classification, a measure of how well our procedure works is in order. We have provided simple functions to calculate the median F-measure (see @optimalFlow for details). 
```{r ej85, echo = TRUE}
scoreF1.optimalFlow <- optimalFlow::f1Score(classification.optimalFlow$cluster,
                                            test.cytometry, noise.types)
print(scoreF1.optimalFlow)
```
We see that median F1-score, the first row in the table is close to 1 for each cell type, reflecting that classification is good for each cell type.

When using a consensus method that is not pooling, the default in optimalFlowTemplates, we have to assign cell types to the clusters in the prototype cytometries. This is done by voting, since our database is formed by gated cytometries and we have assigned cell types.
```{r ej9, echo = TRUE}
classification.optimalFlow.barycenter <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow.barycenter, consensus.method = "k-barycenter", cl.paral = 1
    )
```

```{r ej92, echo = TRUE}
table(classification.optimalFlow.barycenter$cluster)
classification.optimalFlow.barycenter$cluster.vote
```
In this case the fuzzy classification is a hard one, since values of simple.proportion are all one. This means that label 1 is assigned to the entry in cell and so on.
```{r ej922, echo = TRUE}
classification.optimalFlow.barycenter$assigned.template.index
templates.optimalFlow.barycenter$clustering
```
Since usually the relabelling in classification.optimalFlow.barycenter$cluster.vote is fuzzy, we need to convert it to a hard clustering and then apply our median F-measure criteria. This is done as follows.
```{r ej93, echo = TRUE}
scoreF1.optimalFlow.barycenter <- 
  f1ScoreVoting(
    classification.optimalFlow.barycenter$cluster.vote, classification.optimalFlow.barycenter$cluster,
    test.cytometry,
    1.01, noise.types
    )
print(scoreF1.optimalFlow.barycenter$F1_score)
```

Exactly the same applies for the case of templates.optimalFlow.hdbscan.
```{r ej10, echo = TRUE}
classification.optimalFlow.hdbscan <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow.hdbscan, consensus.method = "hierarchical", cl.paral = 1
    )
table(classification.optimalFlow.hdbscan$cluster)
classification.optimalFlow.hdbscan$cluster.vote
classification.optimalFlow.hdbscan$assigned.template.index
templates.optimalFlow.hdbscan$clustering
scoreF1.optimalFlow.hdbscan <-
  f1ScoreVoting(
    classification.optimalFlow.hdbscan$cluster.vote, classification.optimalFlow.hdbscan$cluster,
   test.cytometry,
    1.01, noise.types
  )
print(scoreF1.optimalFlow.hdbscan$F1_score)
```
Another way of doing classification is to relabel the initial clustering that we obtain using the most similar template obtained by optimalFlowTemplates. This is called label-transfer and is further explained in @optimalFlow.
```{r ej11, echo = TRUE}
classification.optimalFlow.2 <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow, consensus.method = "pooling", classif.method = "matching",
    cost.function = "ellipses", cl.paral = 1
    )
table(classification.optimalFlow.2$cluster)
table(classification.optimalFlow.2$clusterings[[1]]$cluster)
classification.optimalFlow.2$cluster.vote
classification.optimalFlow.2$assigned.template.index
templates.optimalFlow$clustering
scoreF1.optimalFlow.2 <-
  f1ScoreVoting(
    classification.optimalFlow.2$cluster.vote, classification.optimalFlow.2$cluster,
    test.cytometry,
    1.01, noise.types
    )
print(scoreF1.optimalFlow.2$F1_score)
```
Also, classical techniques as random forest are available.
```{r ej12, echo = TRUE}
classification.optimalFlow.3 <-
  optimalFlowClassification(
    test.cytometry[, 1:10],
    database, templates.optimalFlow, consensus.method = "pooling",
    classif.method = "random forest", cl.paral = 1
    )
table(classification.optimalFlow.3$cluster)
classification.optimalFlow.3$assigned.template.index # the cytometry used for learning belongs to the cluster labelled as 1 and is the first of the cytometries in that cluster, hence it is the first cytometry in the database.
templates.optimalFlow$clustering
scoreF1.optimalFlow.3 <-
  optimalFlow::f1Score(classification.optimalFlow.3$cluster,
    test.cytometry,
    noise.types
    )
print(scoreF1.optimalFlow.3)
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimalFlowClassification.R
\name{optimalFlowClassification}
\alias{optimalFlowClassification}
\title{optimalFlowClassification}
\usage{
optimalFlowClassification(
  X,
  database,
  templates,
  consensus.method = "pooling",
  cov.estimation = "standard",
  alpha.cov = 0.85,
  initial.method = "supervized",
  max.clusters = NA,
  alpha.tclust = 0,
  restr.factor.tclust = 1000,
  classif.method = "qda",
  qda.bar = TRUE,
  cost.function = "points",
  cl.paral = 1,
  equal.weights.voting = TRUE,
  equal.weights.template = TRUE
)
}
\arguments{
\item{X}{Datasample to be classified.}

\item{database}{A list where each entry is a partition (clustering) represented as dataframe, of the same dimensions, where the last variable represents the labels of the partition.}

\item{templates}{List of the consensus clusterings for every group in the partition of the database obtained by optimalFlowTemplates}

\item{consensus.method}{The consensus.method value that was used in optimalFlowTemplates.}

\item{cov.estimation}{How to estimate covariance matrices in each cluster of a partition. "standard" is for using cov(), while "robust" is for using robustbase::covMcd.}

\item{alpha.cov}{Only when cov.estimation = "robust". Indicates the value of alpha in robustbase::covMcd.}

\item{initial.method}{Indicates how to obtain a partition of X. Takes values in  c("supervized", "unsupervized"). Supervized uses tclust initilized by templates. Unsupevized usese flowMeans.}

\item{max.clusters}{The maximum numbers of clusters for flowMeans. Only when initial.method = unsupervized.}

\item{alpha.tclust}{Level of trimming allowed fo tclust. Only when initial.method = supervized.}

\item{restr.factor.tclust}{Fixes the restr.fact parameter in tclust. Only when initial.method = supervized.}

\item{classif.method}{Indicates what type of supervised learning we want to do. Takes values on c("matching", "qda", "random forest").}

\item{qda.bar}{Only if classif.method = "qda". If True then the appropriate consensus clustering (template, prototype) is used for learning. If False, the closest partition in the appropriate group is used.}

\item{cost.function}{Only if classif.method = "matching". Indicates the cost function, distance between clusters, to be used for label matching.}

\item{cl.paral}{Number of cores to be used in parallel procedures.}

\item{equal.weights.voting}{only when classif.method = "qda" and qda.bar =F, or when  classif.method = "random forest". Indicates the weights structure when looking for the most similar partition in a group.}

\item{equal.weights.template}{If True, weights assigned to every cluster in a partion are uniform (1/number of clusters). If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.}
}
\value{
A list formed by:
\describe{
\item{cluster}{Labels assigned to the input data.}
\item{clusterings}{A list that contains the initial unsupervized or semi-supervized clusterings of the cytometry of interest. Can have as much entries as the number of templates in the semi-supervized case (initial.method = "supervized), or only one entry in the case of initial.method = "unsupervized". Each entry is a list where the most relevant argument for the clusterings is cluster.}
\item{assigned.template.index}{Label of the group for which the template is closer to the data. When classical qda or random forest ares used for classification there is a secon argument indicating the index of the cytometry in the cluster used for learning.}
\item{cluster.vote}{Only when classif.method = "matching" or when consensus.method in c("hierarchical", "k-barycenter"). Vote on the type of every label in the partition of the data. In essence, cluster + cluster.vote return a fuzzy clustering of the data of interest.}
}
}
\description{
Performs a supervised classification of input data when a database and a partition of the database are provided.
}
\examples{
# # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
database <- buildDatabase(
  dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
    population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
# # To select the appropriate number of templates, via hierarchical tree, in an interactive fashion and produce a clustering we can also use:
# templates.optimalFlow <- optimalFlowTemplates(database = database)
templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
                                             cl.paral = 1)
classification.optimalFlow <- optimalFlowClassification(Cytometry1[
  which(match(Cytometry1$`Population ID (name)`,c("Monocytes", "CD4+CD8-", "Mature SIg Kappa",
                                                  "TCRgd-"), nomatch = 0) > 0), 1:10], database, templates.optimalFlow, cl.paral = 1)
scoreF1.optimalFlow <- optimalFlow::f1Score(classification.optimalFlow$cluster,
                                           Cytometry1[which(match(Cytometry1$`Population ID (name)`,
                                                                                 c("Monocytes", "CD4+CD8-", "Mature SIg Kappa", "TCRgd-"), nomatch = 0) > 0),], noise.types)


}
\references{
E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f1ScoreVoting.R
\name{f1ScoreVoting}
\alias{f1ScoreVoting}
\title{f1ScoreVoting}
\usage{
f1ScoreVoting(voting, clustering, cytometry, nivel_sup, noise.cells)
}
\arguments{
\item{voting}{A list where each entry is a vote on the respective label.}

\item{clustering}{Labels of the partition.}

\item{cytometry}{Data of the clustering, where the last variable contains the original labels.}

\item{nivel_sup}{level of tolerance for assigning a hard clustering. Should be greater or equal than 1. Class A is assigned if class A > nivel_sup * Class B.}

\item{noise.cells}{An array of labels to be considered as noise.}
}
\value{
A matrix where the first row is the F1 score, the second row is the Precision and the third row is the Recall.
}
\description{
Calculates the F1 score fore each group in a partition, when provided with a fuzzy classification.
}
\examples{
# # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
database <- buildDatabase(
  dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
    population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))

templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
cl.paral = 1)

classification.optimalFlow <- optimalFlowClassification(as.data.frame(Cytometry1)[
 which(match(Cytometry1$`Population ID (name)`, c('Monocytes', 'CD4+CD8-',
                                                 'Mature SIg Kappa', 'TCRgd-'), nomatch = 0) > 0), 1:10], database, templates.optimalFlow,
 classif.method = 'matching', cost.function = 'ellipses', cl.paral = 1)

f1ScoreVoting(classification.optimalFlow$cluster.vote, classification.optimalFlow$cluster,
              as.data.frame(Cytometry1)[which(match(Cytometry1$`Population ID (name)`,
                                                    c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'), nomatch = 0) > 0), ], 1.01, noise.types)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimationCellBarycenter.R
\name{estimationCellBarycenter}
\alias{estimationCellBarycenter}
\title{estimationCellBarycenter}
\usage{
estimationCellBarycenter(cell, cytometries)
}
\arguments{
\item{cell}{Name of the cluster of interest.}

\item{cytometries}{List of clusterings.}
}
\value{
A list representing the (1-)barycenter:
\describe{
 \item{mean}{Mean of the barycenter.}
 \item{cov}{Covariance of the barycenter.}
 \item{weight}{Weight associated to the barycenter.}
 \item{type}{Type of the cluster.}
}
}
\description{
Estimates a Wasserstein barycenter for a cluster type using a collection of partitions.
}
\examples{
partition1 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = '1'),
                  list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = '2'))
partition2 <- list(list(mean = c(1, -1), cov = diag(1, 2), weight = 0.5, type = '1'),
                  list(mean = c(-1, 1), cov = diag(1, 2), weight = 0.5, type = '2'))
cytometries <- list(partition1, partition2)
estimationCellBarycenter('1',cytometries)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tclustWithInitialization.R
\name{tclustWithInitialization}
\alias{tclustWithInitialization}
\title{tclustWithInitialization}
\usage{
tclustWithInitialization(
  initialization,
  cytometry,
  i.sol.type = "points",
  trimming = 0.05,
  restr.fact = 1000
)
}
\arguments{
\item{initialization}{Initial solution for parameters provided by the user. Can be a matrix of data containing observations anc cluster assignations or can be a list spesifying a multivariate mixture of gaussians.}

\item{cytometry}{A matrix or data.frame of dimension n x p, containing the observations (row-wise).}

\item{i.sol.type}{Type of initial solutions in c('points', 'barycenters'). 'points' refers to a classified data matrix, while 'barycenters' to a multivariate mixture.}

\item{trimming}{The proportion of observations to be trimmed.}

\item{restr.fact}{The constant restr.fact >= 1 constrains the allowed differences among group scatters. Larger values imply larger differences of group scatters, a value of 1 specifies the strongest restriction.}
}
\value{
A list with entries:
\describe{
 \item{cluster}{A numerical vector of size n containing the cluster assignment for each observation. Cluster names are integer numbers from 1 to k, 0 indicates trimmed observations.}
 \item{n_clus}{Number of clusters actually found.}
 \item{obj}{he value of the objective function of the best (returned) solution.}
}
}
\description{
A wrapper for the function tclust_H.
}
\examples{
x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2),
        matrix(rnorm(100) + 4, ncol = 2))
## robust cluster obtention from a sample x asking for 3 clusters,
## trimming level 0.05 and constrain level 12
k <- 3; alpha <- 0.05; restr.fact <- 12
output = tclust_H(x = x, k = k, alpha = alpha, nstart = 50, iter.max = 20,
                 restr = 'eigen', restr.fact = restr.fact, sol_ini_p = FALSE, sol_ini = NA,
                 equal.weights = FALSE, trace = 0, zero.tol = 1e-16)
## cluster assigment
output2 <- tclustWithInitialization(data.frame(x, output$cluster), x, 'points', 0.05, 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cytoPlot3d.R
\name{cytoPlot3d}
\alias{cytoPlot3d}
\title{cytoPlot3d}
\usage{
cytoPlot3d(
  cytometry.as.mixture,
  dimensions = c(1, 2),
  xlim = NULL,
  ylim = NULL,
  zlim = NULL,
  xlab = NULL,
  ylab = NULL,
  zlab = NULL
)
}
\arguments{
\item{cytometry.as.mixture}{A list, where each element contains the parameters of a component of the mixture as a list with entries: mean, cov, weight and type.}

\item{dimensions}{A vector containing the three variables on which to perform the projection.}

\item{xlim}{the x limits (x1, x2) of the plot. Note that x1 > x2 is allowed and leads to a ‘reversed axis’. The default value, NULL, indicates that the range of the finite values to be plotted should be used.}

\item{ylim}{the y limits of the plot.}

\item{zlim}{the z limits of the plot.}

\item{xlab}{a label for the x axis, defaults to a description of x.}

\item{ylab}{a label for the y axis, defaults to a description of y.}

\item{zlab}{a label for the z axis, defaults to a description of z.}
}
\value{
A three dimensional plot of ellipsoids containing the 95% of probability for the respectives components of the mixture distribution.
}
\description{
A rgl::plot3d wrapper for cytometries as a mixture of multivariate normals as used in optimalFlowTemplates.
}
\examples{
database <- buildDatabase(
  dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
  population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
templates.optimalFlow <-
  optimalFlowTemplates(
    database = database, templates.number = 5, cl.paral = 1
  )
# # To execute requires an actual monitor since it uses rgl.
# cytoPlot3d(templates.optimalFlow$templates[[3]], dimensions = c(4, 3, 9), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000), xlab = "", ylab = "", zlab = ")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tclustAuxiliaryFunctions_dd.R
\name{tclust_H}
\alias{tclust_H}
\title{tclust_H}
\usage{
tclust_H(
  x,
  k = 3,
  alpha = 0.05,
  nstart = 50,
  iter.max = 20,
  restr = "eigen",
  restr.fact = 12,
  sol_ini_p = FALSE,
  sol_ini = NA,
  equal.weights = FALSE,
  trace = 0,
  zero.tol = 1e-16
)
}
\arguments{
\item{x}{A matrix or data.frame of dimension n x p, containing the observations (row-wise).}

\item{k}{The number of clusters initially searched for.}

\item{alpha}{The proportion of observations to be trimmed.}

\item{nstart}{The number of random initializations to be performed. Only when sol_ini_p = FALSE.}

\item{iter.max}{The maximum number of concentration steps to be performed. The concentration steps are stopped, whenever two consecutive steps lead to the same data partition.}

\item{restr}{The type of restriction to be applied on the cluster scatter matrices. Valid values are "eigen" (default).}

\item{restr.fact}{The constant restr.fact >= 1 constrains the allowed differences among group scatters. Larger values imply larger differences of group scatters, a value of 1 specifies the strongest restriction.}

\item{sol_ini_p}{Initial solution for parameters provided by the user TRUE/FALSE, if TRUE is stored in sol_ini.}

\item{sol_ini}{Initial solution for parameters provided by the user.}

\item{equal.weights}{A logical value, specifying whether equal cluster weights (TRUE) or not (FALSE) shall be considered in the concentration and assignment steps.}

\item{trace}{Defines the tracing level, which is set to 0 by default. Tracing level 2 gives additional information on the iteratively decreasing objective function's value.}

\item{zero.tol}{The zero tolerance used. By default set to 1e-16.}
}
\value{
A list with values:
\describe{
 \item{centers}{A matrix of size p x k containing the centers (column-wise) of each cluster.}
 \item{cov}{An array of size p x p x k containing the covariance matrices of each cluster.}
 \item{cluster}{A numerical vector of size n containing the cluster assignment for each observation. Cluster names are integer numbers from 1 to k, 0 indicates trimmed observations.}
 \item{par}{A list, containing the parameters the algorithm has been called with (x, if not suppressed by store.x = FALSE, k, alpha, restr.fact, nstart, KStep, and equal.weights).}
 \item{weights}{A numerical vector of length k, containing the weights of each cluster.}
 \item{obj}{he value of the objective function of the best (returned) solution.}
}
}
\description{
A wrapper for the internal fucntion tclust_. Performs robust non spherical clustering, tclust, where initial solutions are allowed.
}
\details{
This iterative algorithm initializes k clusters randomly and performs "concentration steps" in order to improve the current cluster assignment. The number of maximum concentration steps to be performed is given by iter.max. For approximately obtaining the global optimum, the system is initialized nstart times and concentration steps are performed until convergence or iter.max is reached. When processing more complex data sets higher values of nstart and iter.max have to be specified (obviously implying extra computation time). However, if more then half of the iterations would not converge, a warning message is issued, indicating that nstart has to be increased.

The parameter restr defines the cluster's shape restrictions, which are applied on all clusters during each iteration. Options "eigen"/"deter" restrict the ratio between the maximum and minimum eigenvalue/determinant of all cluster's covariance structures to parameter restr.fact. Setting restr.fact to 1, yields the strongest restriction, forcing all eigenvalues/determinants to be equal and so the method looks for similarly scattered (respectively spherical) clusters. Option "sigma" is a simpler restriction, which averages the covariance structures during each iteration (weighted by cluster sizes) in order to get similar (equal) cluster scatters.
}
\examples{
x <- rbind(matrix(rnorm(100), ncol = 2), matrix(rnorm(100) + 2, ncol = 2),
        matrix(rnorm(100) + 4, ncol = 2))
## robust cluster obtention from a sample x asking for 3 clusters,
## trimming level 0.05 and constrain level 12
k <- 3; alpha <- 0.05; restr.fact <- 12
output <- tclust_H(x = x, k = k, alpha = alpha, nstart = 50, iter.max = 20,
                 restr = "eigen", restr.fact = restr.fact, sol_ini_p = FALSE, sol_ini = NA,
                 equal.weights = FALSE, trace = 0, zero.tol = 1e-16)
## cluster assigment
output$cluster
plot(x, col = output$cluster)

}
\references{
Fritz, H., Garcia-Escudero, L. A., & Mayo-Iscar, A. (2012). tclust: An r package for a trimming approach to cluster analysis. Journal of Statistical Software, 47(12), 1-26.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qdaClassification.R
\name{qdaClassification}
\alias{qdaClassification}
\title{qdaClassification}
\usage{
qdaClassification(normal, data)
}
\arguments{
\item{normal}{A list with arguments mean, covaruance and weight.}

\item{data}{Data frame or matrix on which to perform qda.}
}
\value{
A score for each point.
}
\description{
Gives quadratic discriminant scores to the points in data for a multivariate normal.
}
\examples{
data.qda = cbind(rnorm(50), rnorm(50))
exp(qdaClassification(list(mean = c(0,0), cov = diag(1,2), weight = 1), data.qda))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimCovCellGeneral.R
\name{estimCovCellGeneral}
\alias{estimCovCellGeneral}
\title{estimCovCellGeneral}
\usage{
estimCovCellGeneral(cell, cytometry, labels, type = "standard", alpha = 0.85)
}
\arguments{
\item{cell}{Labell of the clsuter of interest.}

\item{cytometry}{Data of the partition, without labels.}

\item{labels}{Labels of the partition.}

\item{type}{How to estimate covariance matrices of a cluster. 'standard' is for using cov(), while 'robust' is for using robustbase::covMcd.}

\item{alpha}{Only when type = 'robust'. Indicates the value of alpha in robustbase::covMcd.}
}
\value{
A list containing:
\describe{
 \item{mean}{Mean of the cluster.}
 \item{cov}{Covariance of the cluster.}
 \item{weight}{Weight associated to the cluster.}
 \item{type}{Type of the cluster.}
}
}
\description{
Estimation of mean and covariance for a label in a partition.
}
\examples{
estimCovCellGeneral('Basophils', Cytometry1[,1:10], Cytometry1[,11])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/costWasserMatchingEllipse.R
\name{costWasserMatchingEllipse}
\alias{costWasserMatchingEllipse}
\title{costWasserMatchinEllipse}
\usage{
costWasserMatchingEllipse(
  test.cytometry,
  training.cytometries,
  equal.weights = FALSE
)
}
\arguments{
\item{test.cytometry}{A clusetering represented as a list of clusters. Each cluster is a list with elements mean, cov, weight and type.}

\item{training.cytometries}{A list of clusterings with the same format as test.cytometry.}

\item{equal.weights}{If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.}
}
\value{
A vector representing the similarity distance between test.cytometry and the elements in training.cytometries.
}
\description{
Calculates a similarity distance based on the 2-Wassertein distance between mixtures of multivariate normal distributions.
}
\examples{
partition1 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = '1'),
                  list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = '2'))
partition2 <- list(list(list(mean = c(1, -1), cov = diag(1, 2),
                            weight = 0.5, type = '1'), list(mean = c(-1, 1), cov = diag(1, 2), weight = 0.5, type = '2')))
costWasserMatchingEllipse(partition1, partition2)

}
\references{
E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/w2dist.R
\name{w2dist}
\alias{w2dist}
\title{w2dist}
\usage{
w2dist(P, Q)
}
\arguments{
\item{P}{A multivariate normal distribution given as a list with arguments mean and cov.}

\item{Q}{A multivariate normal distribution given as a list with arguments mean and cov.}
}
\value{
A double giving the 2-Wasserstein distance between the two distributions.
}
\description{
The 2-Wasserstein distance between two multivariate normal distributions
}
\examples{
P <- list(mean = c(1, 1), cov = diag(1, 2))
Q <- list(mean = c(0, 0), cov = 1.1*diag(1, 2))
Q <- list(mean = c(0, 0), cov = 1.1*diag(1, 2))
w2dist(P, Q)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f1Score.R
\name{f1Score}
\alias{f1Score}
\title{f1Score}
\usage{
f1Score(clustering, cytometry, noise.cells)
}
\arguments{
\item{clustering}{The labels of the new classification.}

\item{cytometry}{Data of the clustering, where the last variable contains the original labels.}

\item{noise.cells}{An array of labels to be considered as noise.}
}
\value{
A matrix where the first row is the F1 score, the second row is the Precision and the third row is the Recall.
}
\description{
Calculates the F1 score fore each group in a partition.
}
\examples{
f1Score(dplyr::pull(Cytometry3[c(sample(1:250,250),251:(dim(Cytometry3)[1])),],11),
        Cytometry3, noise.types)

}
\references{
E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labelTransfer.R
\name{labelTransfer}
\alias{labelTransfer}
\title{labelTransfer}
\usage{
labelTransfer(
  training.cytometry,
  test.cytometry,
  test.partition,
  equal.weights = FALSE
)
}
\arguments{
\item{training.cytometry}{List of partitions, where each partition is a dataframe where the last column contains the labels of the partition.}

\item{test.cytometry}{Test data, a dataframe without labels.}

\item{test.partition}{Labels of a partition of the test data.}

\item{equal.weights}{If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.}
}
\value{
A fuzzy relabeling consistent of a transportation plan.
}
\description{
Label transfer between a test partition and a training set of partitions.
}
\examples{
\donttest{
data.example <- data.frame(v1 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)),
                          v2 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)), id = c(rep(0, 50), rep(1, 50)))
test.labels <- c(rep('a', 50), rep('b', 50))
labelTransfer(data.example, data.example[, 1:2], test.labels)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/voteLabelTransfer.R
\name{voteLabelTransfer}
\alias{voteLabelTransfer}
\title{voteLabelTransfer}
\usage{
voteLabelTransfer(
  type = "points",
  test.partition,
  test.cytometry,
  test.partition.ellipse,
  training.cytometries,
  training.cytometries.barycenter,
  test = 1,
  op.syst,
  cl.paral = 1,
  equal.weights = FALSE
)
}
\arguments{
\item{type}{'points' indicates use of labelTransfer; 'ellipses' of labelTransferEllipse.}

\item{test.partition}{Only when type = 'points'. Labels of a partition of the test data.}

\item{test.cytometry}{Only when type = 'points'. Test data, a dataframe without labels.}

\item{test.partition.ellipse}{Only when type = 'ellipses'. A test clustering viewed as a mixture of multivariate normal distributions.}

\item{training.cytometries}{Only when type = 'points'. List of partitions, where each partition is a dataframe wher the last column contains the labels of the partition.}

\item{training.cytometries.barycenter}{Only when type = 'ellipses'. A training partition viewed as a mixture of multivariate normal distributions.}

\item{test}{Only when type = 'ellipses'. A dummy variable, should be any integral. Ment for use with lapply.}

\item{op.syst}{Type of system, takes values in c('unix', 'windows').}

\item{cl.paral}{Number of cores to be used in parallel procedures.}

\item{equal.weights}{If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.}
}
\value{
A list containing:
\describe{
 \item{final.vote}{A list for the votes on each cell.}
 \item{complete.vote}{A more complete list for the votes on each cell.}
}
}
\description{
A wrapper for doing either labelTransfer or labelTransferEllipse.
}
\examples{
\donttest{
data.example <- data.frame(v1 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)),
                          v2 = c(rnorm(50, 2, 1), rnorm(50, -2, 1)), id = c(rep(0, 50), rep(1, 50)))
test.labels <- c(rep('a', 50), rep('b', 50))
voteLabelTransfer(test.partition = test.labels, test.cytometry = data.example[, 1:2],
                  training.cytometries = list(data.example), op.syst = .Platform$OS.type)$final.vote[[1]]
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labelTransferEllipse.R
\name{labelTransferEllipse}
\alias{labelTransferEllipse}
\title{labelTransferEllipse}
\usage{
labelTransferEllipse(
  i,
  test.cytometry.ellipses,
  training.cytometries.barycenter,
  equal.weights = FALSE
)
}
\arguments{
\item{i}{A dummy variable, should be any integral. Ment for use with lapply.}

\item{test.cytometry.ellipses}{A test clustering viewed as a mixture of multivariate normal distributions.}

\item{training.cytometries.barycenter}{A training partition viewed as a mixture of multivariate normal distributions.}

\item{equal.weights}{If True, weights assigned to every cluster in a partion are uniform (1/number of clusters) when calculating the similarity distance. If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.}
}
\value{
A fuzzy relabeling consistent of a transportation plan.
}
\description{
Label transfer between a test partition and a training partitions viewed as a mixture of gaussians.
}
\examples{
partition1 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = '1'),
                  list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = '2'))
partition2 <- list(list(mean = c(1, 1), cov = diag(1, 2), weight = 0.5, type = 'a'),
                  list(mean = c(-1, -1), cov = diag(1, 2), weight = 0.5, type = 'b'))
labelTransferEllipse(1, partition2, partition1)
}
\references{
E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimalFlowTemplates.R
\name{optimalFlowTemplates}
\alias{optimalFlowTemplates}
\title{optimalFlowTemplates}
\usage{
optimalFlowTemplates(
  database,
  database.names = NULL,
  cov.estimation = "standard",
  alpha.cov = 0.85,
  equal.weights.template = TRUE,
  hclust.method = "complete",
  trimm.template = FALSE,
  templates.number = NA,
  minPts = 2,
  eps = 1,
  consensus.method = "pooling",
  barycenters.number = NA,
  bar.repetitions = 40,
  alpha.bar = 0.05,
  bar.ini.method = "plus-plus",
  consensus.minPts = 3,
  cl.paral = 1
)
}
\arguments{
\item{database}{A list where each entry is a partition (clustering) represented as dataframe, of the same dimensions, where the last variable represents the labels of the partition.}

\item{database.names}{Names of the elements in the database.}

\item{cov.estimation}{How to estimate covariance matrices in each cluster of a partition. 'standard' is for using cov(), while 'robust' is for using robustbase::covMcd.}

\item{alpha.cov}{Only when cov.estimation = 'robust'. Indicates the value of alpha in robustbase::covMcd.}

\item{equal.weights.template}{If True, weights assigned to every cluster in a partion are uniform (1/number of clusters). If False, weights assigned to clusters are the proportions of points in every cluster compared to the total amount of points in the partition.}

\item{hclust.method}{Indicates what kind of hierarchical clustering to  do with the similarity distances matrix of the partitions. Takes values in c('complete', 'single', 'average', 'hdbscan', 'dbscan').}

\item{trimm.template}{Logical value. Indicates if it is allowed to not take into account some of the entries of database. Default is False.}

\item{templates.number}{Only if hclust.method in c('complete', 'single', 'average'). Indicates the number of clusters to use with cutree. If set to NA (default), plots the hierarchical tree and asks the user to introduce an appropriate number of clusters.}

\item{minPts}{Only if hclust.method in c('hdbscan', 'dbscan'). Indicates the value of argument minPts in dbscan::dbscan and dbscan::hdbscan.}

\item{eps}{Only if hclust.method = 'dbscan'. Indicates the value of eps in dbscan::dbscan.}

\item{consensus.method}{Sets the way of doing consensus clustering when clusters are viewed as Multivariate Distributions. Can take values in c('pooling', 'k-barycenter', 'hierarchical'). See details.}

\item{barycenters.number}{Only if consensus.method = 'k-barycenter'. Sets the number, k, of barycenters when using k-barycenters.}

\item{bar.repetitions}{Only if consensus.method = 'k-barycenter'. How many times to repeat the k-barycenters procedure. Equivalent to nstart in kmeans.}

\item{alpha.bar}{Only if consensus.method = 'k-barycenter'. The level of trimming allowed during the k-barycenters procedure.}

\item{bar.ini.method}{Only if consensus.method = 'k-barycenter'. Takes values in c('rnd', 'plus-plus'). See details.}

\item{consensus.minPts}{Only if consensus.method = 'hierarchical'. The value of argument minPts for dbscan::hdbscan.}

\item{cl.paral}{Number of cores to be used in parallel procedures.}
}
\value{
A list containting:
\describe{
 \item{templates}{A list representing the consensus clusterings for every group in the partition of the database. Each element of the list is a template partition. Hence it is a list itself, containig the cell types in the prototype, where each element has components: mean, cov, weight and type.}
 \item{clustering}{Clustering of the input partitions.}
 \item{database.elliptical}{A list containig each cytometry in the database viewed as a mixture distribution. Each element of the list is a cytometry viewed as a mixture. Hence it is a list itself, containig the cell types in the cytometry, where each element has components: mean, cov, weight and type.}
}
}
\description{
Returns a partition of the input clusterings with a respective consensus clustering for every group.
}
\examples{
# # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
database <- buildDatabase(
  dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
    population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))

# # To select the appropriate number of templates, via hierarchical tree, in an interactive fashion and produce a clustering we can also use:
# templates.optimalFlow <- optimalFlowTemplates(database = database)

templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
                                             cl.paral = 1)

}
\references{
E del Barrio, H Inouzhe, JM Loubes, C Matran and A Mayo-Iscar. (2019) optimalFlow: Optimal-transport approach to flow cytometry gating and population matching. arXiv:1907.08006
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cytoPlotDatabase3d.R
\name{cytoPlotDatabase3d}
\alias{cytoPlotDatabase3d}
\title{cytoPlotDatabase3d}
\usage{
cytoPlotDatabase3d(
  database.cytometries.as.mixtures,
  dimensions = c(1, 2, 3),
  xlim = c(0, 8000),
  ylim = c(0, 8000),
  zlim = c(0, 8000),
  xlab = "",
  ylab = "",
  zlab = "",
  colour = TRUE
)
}
\arguments{
\item{database.cytometries.as.mixtures}{A list where each component is a mixture distribution. That is, each component is a list, where each element contains the parameters of a component of the mixture as a list with entries: mean, cov, weight and type.}

\item{dimensions}{A vector containing the two variables on which to perform the projection.}

\item{xlim}{the x limits (x1, x2) of the plot. Note that x1 > x2 is allowed and leads to a ‘reversed axis’. The default value, NULL, indicates that the range of the finite values to be plotted should be used.}

\item{ylim}{the y limits of the plot.}

\item{zlim}{the z limits of the plot.}

\item{xlab}{a label for the x axis, defaults to a description of x.}

\item{ylab}{a label for the y axis, defaults to a description of y.}

\item{zlab}{a label for the z axis, defaults to a description of z.}

\item{colour}{If TRUE plots elements of a mixture distribution in different colours. If FALSE plots them in black.}
}
\value{
A three dimensional plot of ellipsoids containing the 95% of probability for the respectives components of the mixture distributions.
}
\description{
A plot3d wrapper for a database (list) of cytometries as a mixture of multivariate normals as used in optimalFlowTemplates.
}
\examples{
database <- buildDatabase(
  dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
  population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))
templates.optimalFlow <-
  optimalFlowTemplates(
    database = database, templates.number = 5, cl.paral = 1
  )
# # To execute requires an actual monitor since it uses rgl.
# cytoPlotDatabase3d(templates.optimalFlow$database.elliptical[which(templates.optimalFlow$clustering == 3)], dimensions = c(4,3), xlim = c(0, 8000), ylim = c(0, 8000), zlim = c(0, 8000), xlab = "", ylab = "", ylab = c(0, 8000))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trimmedKBarycenter.R
\name{trimmedKBarycenter}
\alias{trimmedKBarycenter}
\title{trimmedKBarycenter}
\usage{
trimmedKBarycenter(k, alpha0, type.ini = "rnd", reps.list)
}
\arguments{
\item{k}{Number k of elements in the k-barycenter.}

\item{alpha0}{Level of trimming.}

\item{type.ini}{of initialization in c('rnd', 'plus-plus'). 'rnd' makes the common random
initilaization while 'plus-plus' initializes in a similar fashion to k-means++.}

\item{reps.list}{List of multivariate normals for which the trimmed k-barycenter should be performed.}
}
\value{
A list with values:
\describe{
 \item{variacion_wasser}{A double giving the Waserstein variation.}
 \item{baricentro}{A list of k elements, each of which is a member of the k-barycenter.
  Each eement is a normal distribution characterized by a mean and a covariance.}
 \item{cluster}{The assignment of the original entries to each member of the k-barycenter.}
}
}
\description{
Calculates a 2-Wasserstein k-barycenter of a list of multivariate normal distributions.
}
\examples{
normals <- list(list(mean = c(1, 1), cov = diag(2, 2)), list(mean = c(1, 1),cov = diag(1, 2)),
 list(mean = c(3, 3), cov = diag(1, 2)))
trimmedKBarycenter(2, 0, 'rnd', normals)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wasserCostFunction.R
\name{wasserCostFunction}
\alias{wasserCostFunction}
\title{wasserCostFunction}
\usage{
wasserCostFunction(j, i, cytometries, equal.weights = FALSE)
}
\arguments{
\item{j}{An entry of the list of partitions.}

\item{i}{An entry of the list of partitions.}

\item{cytometries}{The list of partitions.}

\item{equal.weights}{If True, weights assigned to every cluster in a partion are
uniform (1/number of clusters) when calculating the similarity distance.
If False, weights assigned to clusters are the proportions of points in every
 cluster compared to the total amount of points in the partition.}
}
\value{
A double giving the value of the similarity distance.
}
\description{
Calculates the similarity distance between elements j and i of a list of partitions.
}
\examples{
# # We construct a simple database selecting only some of the Cytometries and some cell types for simplicity and for a better visualisation.
database <- buildDatabase(
  dataset_names = paste0('Cytometry', c(2:5, 7:9, 12:17, 19, 21)),
    population_ids = c('Monocytes', 'CD4+CD8-', 'Mature SIg Kappa', 'TCRgd-'))

templates.optimalFlow <- optimalFlowTemplates(database = database, templates.number = 5,
cl.paral = 1)
print(wasserCostFunction(1, 2, list(templates.optimalFlow$database.elliptical[[1]],
 templates.optimalFlow$database.elliptical[[2]])))

}
